#! /usr/bin/env python3
# coding: utf-8
import curses

# function to deal properly with all kind of typed key
# get_wch() returns a str for standard keys;
# for functions keys, sometimes string names, sometimes escaped sequences,
# sometimes ordinal integers.
# stupid function that translates different possibilities into strings
# for the fun, it's terminal-dependant

# roughly tested with urvxt, screen, xterm terminals
# probably doesn't work with yours


def getkey(stdscr, debug=False):
    key = stdscr.get_wch()
    origkey = key
    if key == "":
        # half delay mode of 1/10s: no input key raises an error
        # assumption: no humain being type ESC + real keys so fast
        curses.halfdelay(1)
        try:
            # get the false keys and stack it
            while 1:
                key += str(stdscr.get_wch())
        except:
            # the bundle of special keys stopped:
            # back to normal mode
            curses.cbreak()
            origkey = key
            if key in ["[b", "[1;2B"]:
                key = "kDN"  # shift down
            elif key in ["[a", "[1;2A"]:
                key = "kUP"  # shift right
            elif key in ["[d", "[1;2D"]:
                key = "KEY_SLEFT"
            elif key in ["[c", "[1;2C"]:
                key = "KEY_SRIGHT"
            # non curses character: control+key
            elif key in ["Ob", "[1;5B"]:
                key = "KEY_CDOWN"
            elif key in ["Oa", "[1;5A"]:
                key = "KEY_CUP"
            elif key in ["Od", "[1;5D"]:
                key = "KEY_CLEFT"
            elif key in ["Oc", "[1;5C"]:
                key = "KEY_CRIGHT"
            elif key in ["OP"]:
                key = "KEY_F(1)"
            elif key in ["OQ"]:
                key = "KEY_F(2)"
            elif key in ["OR"]:
                key = "KEY_F(3)"
            elif key in ["OS"]:
                key = "KEY_F(4)"
            elif key in ["[1~", "OH", "[H"]:
                key = "KEY_HOME"
            elif key in ["[4~", "OF", "[F"]:
                key = "KEY_END"
            else:
                if not debug and len(key) > 1:
                    key = ""

    if type(key) == int:
        if key in [410]:
            key = "KEY_RESIZE"
        elif key in [258]:
            key = "KEY_DOWN"
        elif key in [259]:
            key = "KEY_UP"
        elif key in [260]:
            key = "KEY_LEFT"
        elif key in [261]:
            key = "KEY_RIGHT"
        elif key in [262, 362]:
            key = "KEY_HOME"
        elif key in [360, 385]:
            key = "KEY_END"
        elif key in [263]:
            key = "KEY_BACKSPACE"
        elif key in [265]:
            key = "KEY_F(1)"
        elif key in [266]:
            key = "KEY_F(2)"
        elif key in [267]:
            key = "KEY_F(3)"
        elif key in [268]:
            key = "KEY_F(4)"
        elif key in [269]:
            key = "KEY_F(5)"
        elif key in [270]:
            key = "KEY_F(6)"
        elif key in [271]:
            key = "KEY_F(7)"
        elif key in [272]:
            key = "KEY_F(8)"
        elif key in [273]:
            key = "KEY_F(9)"
        elif key in [274]:
            key = "KEY_F(10)"
        elif key in [275]:
            key = "KEY_F(11)"
        elif key in [276]:
            key = "KEY_F(12)"
        elif key in [330]:
            key = "KEY_DC"
        elif key in [331]:
            key = "KEY_IC"
        elif key in [338]:
            key = "KEY_NPAGE"
        elif key in [339]:
            key = "KEY_PPAGE"
        elif key in [353]:
            key = "KEY_BTAB"
        elif key in [513, 336]:
            key = "kDN"
        elif key in [529, 337]:
            key = "kUP"
        elif key in [393]:
            key = "KEY_SLEFT"
        elif key in [402]:
            key = "KEY_SRIGHT"
        # non curses character: control+key
        elif key in [514, 525]:
            key = "KEY_CDOWN"
        elif key in [530, 566]:
            key = "KEY_CUP"
        elif key in [523, 545]:
            key = "KEY_CLEFT"
        elif key in [528, 560]:
            key = "KEY_CRIGHT"
        elif key in [409]:
            key = "KEY_MOUSE"
        else:
            if not debug:
                key = ""
            else:
                key = str(key)
    return key

## comment previous line and uncomment the followings to use the standalone
#    if debug:
#        return key, origkey
#    else:
#        return key
#
## standalone function
## useful to complete the lists
##import curses
#def main(stdscr):
#    curses.noecho()
#    mmask = (curses.BUTTON1_PRESSED + curses.BUTTON1_RELEASED
#             + curses.BUTTON1_CLICKED + curses.BUTTON1_DOUBLE_CLICKED
#             + curses.BUTTON1_TRIPLE_CLICKED + curses.BUTTON_CTRL)
#    curses.mousemask(mmask)
#
#    ymax, xmax = stdscr.getmaxyx()
#    k = ""
#    i = 0
#    while k != "q":
#        k, ok = getkey(stdscr, debug=True)
#        stdscr.addstr(i%ymax, 0, "                       ")
#        stdscr.addstr(i%ymax, 0, str(k))
#        stdscr.addstr(i%ymax, 15, str(ok))
#        i += 1
#curses.wrapper(main)



#! /usr/bin/env python3
# coding: utf-8
import curses
import datetime
from . import calendar, inputs, culdav, getkey


class Culendar:
    def __init__(self, screen, conf, day, cal, caldav, todo):
        self._screen = screen
        self.conf = conf  # use the specific setter

        self._xoffset = 6  # left offset to draw hours
        self._yoffsettop = 3  # top offset to draw the header
        self._yoffsetbot = 4  # bottom offset to draw the footer

        self._day = day
        self._prevday = datetime.date(1, 1, 1)

        self._cal = cal
        self._caldav = caldav
        self._todo = todo
        # self._calweek is first computed in draw_cal()→draw_week()
        # self._todocol is todowidth in columns, computed in update()

        self._ehl = 0  # highlighted event of current day
        self._prevehl = 0  # idem, previous one
        # highlighted todo, if any, index starts at 1
        self._todohl = 0  # focus on calendar on start
        self._count = 0  # repetitions of the next command
        # compute useful stuff
        self.update()

###############################################################################
#       access and mutate functions
###############################################################################

    @property
    def screen(self):
        return self._screen

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, c):
        self._conf = c
        self._hmin = c.hmin
        self._hmax = c.hmax
        self._daynb = 5 + 2*c.WE  # + if WE included

    @property
    def day(self):
        return self._day

    @day.setter
    def day(self, d):
        self._prevday = self._day
        self.changehl(-self._ehl)  # resets highlight
        self._day = d

    @property
    def cal(self):
        return self._cal

    @cal.setter
    def cal(self, c):
        self._cal = c

    @property
    def caldav(self):
        return self._caldav

    @property
    def hmin(self):
        return self._hmin

    @property
    def hmax(self):
        return self._hmax

    @property
    def hour_size(self):
        return self._hour_size

    @property
    def ymax(self):
        return self._ymax

    @property
    def lines(self):
        return self._lines

    @property
    def daynb(self):
        return self._daynb

    @property
    def day_size(self):
        return self._day_size

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        self._count = count

###############################################################################
#       update functions
###############################################################################
    def dead_duck(self):
        self.clear_cal()
        text = _("\_x< A dead duck. That's all such a small terminal deserves.")
        self._screen.addstr(0, 0, text[0:(self._width-1)*(self._ymax-1)])

        key = ""
        while not((key in self.conf.keys['quit']) or (key == 'KEY_RESIZE')):
            key = self.getkey()
        if key in self.conf.keys['quit']:
            exit()
        else:
            self.update()

    def update(self):
        self._ymax, self._xmax = self._screen.getmaxyx()
        if self.conf.mouse:  # limit to clickable without bug
            self._xmax = min(self._xmax, 223)

        self._caldav.resize()  # update the ymax and xmax in Caldav
        if self.conf.todo:  # keep space for a todo
            xmax = self._xmax - round(self._xmax*self.conf.todowidth)
            self._day_size = (xmax-self._xoffset) // self._daynb
            # update xmax to be hour place + a multiple of day_size
            xmax = 7 + self._day_size * self._daynb
            self._todocol = self._xmax - xmax
            self._xmax = xmax
        else:
            self._todocol = 0
            self._day_size = (self._xmax-self._xoffset) // self._daynb
        self._width = self._xmax + self._todocol
        self._hour_size = (self._ymax - self._yoffsettop - self._yoffsetbot
                           ) // (self._hmax - self._hmin)
        self.linepos()

        if self.conf._debug:
            self.clear_cal()
            self.draw_cal()
        else:
            try:
                # everything changes, clear screen and redraw stuff
                self.clear_cal()
                self.draw_cal()
            except:
                # culendar couldn't draw it: it's ridiculously small terminal
                self.dead_duck()

    def linepos(self):
        # locations of [day lines, hour lines]
        self._lines = [[], []]

        self._lines[0].append(self._xoffset)
        for id in range(self._daynb-1):
            self._lines[0].append(self._lines[0][id] + self._day_size)
        # the last one is in the last position
        self._lines[0].append(self._xmax-1)

        self._lines[1].append(self._yoffsettop)
        for ih in range(self._hmax-self._hmin):
            self._lines[1].append(self._lines[1][ih] + self._hour_size)

    def toggle_WE(self):
        if self._daynb == 5:
            self._daynb = 7
        elif self._daynb == 7:
            self._daynb = 5
            while self._day.isoweekday() in [6, 7]:
                self._day = datetime.date.fromordinal(
                                self._day.toordinal() - 1)
        self.update()

    def avoid_WE(self, shift):
        if (self._day.isoweekday() in [6, 7]) & (self._daynb == 5):
            while self._day.isoweekday() in [6, 7]:
                if shift > 0:
                    self._day = datetime.date.fromordinal(
                                    self._day.toordinal() + 1)
                else:
                    self._day = datetime.date.fromordinal(
                                    self._day.toordinal() - 1)

    def toggle_TODO(self):
        # if todohl > 1, Todo bar is in highlight mode, remove it
        if self._todohl > 0:
            self.conf.todo = False
            self._todohl = 0
            self.update()  # recompute everything
        elif self.conf.todo:  # Todo bar is unhighlighted
            self._todohl = 1
            self.draw_day()  # remove the day hl
            self.draw_header()  # remove the week hl
            self.draw_hl()  # hl the Todo item
        else:  # make appear the Todo bar
            self.conf.todo = True
            self._todohl = 1
            self.update()  # recompute everything

    def addyear(self, shift):
        try:
            daytmp = self.day.replace(year=self.day.year + shift)
            delta = daytmp - self.day
            self.addday(delta.days)  # use the proper function
        except:     # we can't, that's a daytmp is out of range for month
                    # ie, from 19 February to 31 February
            self.addday(shift*365)  # fallback function

    def addmonth(self, shift):
        newmonth = (self.day.month - 1 + shift) % 12 + 1
        try:
            if abs(self.day.month-newmonth) != abs(shift):  # year changed
                self.addyear(int(shift/abs(shift)))  # sign(shift)
            daytmp = self.day.replace(month=newmonth)
            delta = daytmp - self.day
            self.addday(delta.days)  # use the proper function
        except:  # we can't, that's a day is out of range for month
                # ie, from 31 January to 31 February
            self.addday(shift*30)  # fallback function

    def addday(self, shift):
        # change day: autoswitch from todo bar if any
        if self._todohl:
            self._todohl = 0
            self.draw_header()  # hl the weekbar
            self.draw_todo()  # un-hl the todo bar
        # set a new day
        self._prevday = self._day
        self._day = datetime.date.fromordinal(self._day.toordinal()+shift)
        self.avoid_WE(shift)
        # new day, new week?
        if self._day.isocalendar()[1] != self._prevday.isocalendar()[1]:
            # if it's a new week, recompute calweek
            self._calweek = calendar.extract_week(self._cal, self._caldav,
                                                  self._day, self,
                                                  self.conf.autosync)
        # update the highlight, which remembers old values of day and ehl
        self.changehl(0)
        # redraw useful stuff
        prevweek = self._prevday.toordinal() - self._prevday.isoweekday()
        curweek = self._day.toordinal() - self._day.isoweekday()
        self.draw_week()

    def changehl(self, inc):
        if self._todohl:  # Todo bar is highlighted
            maxhl = len(self._todo.events)
            self._todohl = (self._todohl + inc) % (maxhl+1)
            if self._todohl == 0:
                self._todohl = 1  # todo hl starts at index 1
        else:
            self._prevehl = self._ehl
            iday = self._day.isoweekday()-1
            maxhl = len(self._calweek[iday].events)
            if maxhl > 0:
                self._ehl = (self._ehl + inc) % maxhl
            else:
                self._ehl = 0

    def sameday(self):
        self._prevday = self._day

    def changeday(self, day=None):
        self.clear_footer()
        if not day:
            day = self.ask_day()
            if day == "CUL_CANCEL":
                return  # cancel the change of day
        if self.daynb == 5:
            while day.isoweekday() in [6, 7]:
                day = datetime.date.fromordinal(day.toordinal() - 1)
        self.day = day
        self.draw_week()

###############################################################################
#       event functions
###############################################################################

    def del_event(self):
        if self._todohl > 0:
            self.del_todo()
            return
        inform = 0
        # select highlighted event if existing
        iday = self._day.isoweekday()-1
        if len(self._calweek[iday].events) > self._ehl:
            event = self._calweek[iday].events[self._ehl]
            # clear the footer event
            self.clear_footer()
            if event.caldav:
                if event.caldav == "webcal":
                    # webcals are read-only
                    msg = _("Error: can't delete webcal event or offline copy")
                    self._screen.addstr(self._ymax-2, 0, msg)
                    self._screen.addstr(self._ymax-1, 0,
                                        _("Press any key to continue"))
                    key = self.getkey()
                    if key == "KEY_RESIZE":
                        self.update()
                    self.clear_footer()
                else:
                    # delete from local cal in order to
                    # avoid syncing caldav to update events
                    # inform when deleting is slow
                    msg = _("Deleting event from ")
                    url = event.caldav.url.url_raw
                    culdav.del_event(event)
                    self._caldav.events.remove(event)
            else:
                # delete it from the cal
                self._cal.del_event(event)
            # update cal of the week
            self._calweek = calendar.extract_week(self._cal, self._caldav,
                                                  self._day, self,
                                                  autosync=0)
            # redraw day without the deleted event
            self.changehl(0)  # to update the highlighted event

            self.draw_day()
            if inform:
                self.inform(msg+event.url)

    def del_todo(self):
        # reminder: todo_hl starts at index 1
        self._todo.del_event(self._todo.events[self._todohl-1])
        self.changehl(0)  # to update the highlighted event
        self.clear_todo()
        self.draw_todo()

    def edit_event(self):
        if self._todohl > 0:
            self.edit_todo()
            return
        # select highlighted event if existing
        iday = self._day.isoweekday()-1
        if len(self._calweek[iday].events) > self._ehl:
            event = self._calweek[iday].events[self._ehl]

            # we can't edit webcal events
            if event.caldav == "webcal":
                # webcals are read-only
                self.clear_footer()
                msg = _("Error: can't delete webcal event")
                self._screen.addstr(self._ymax-2, 0, msg)
                self._screen.addstr(self._ymax-1, 0,
                                    _("Press any key to continue"))
                key = self.getkey()
                if key == "KEY_RESIZE":
                    self.update()
                self.clear_footer()
                return  # exit here

            # the footer will be useful, clear it
            self.clear_footer()
            # ask question
            self._screen.addstr(self._ymax-2, 0,
  _("Edit: (1) [s]tart time, (2) [e]nd time, (3) [d]escription, (4) [p]lace"))
            s = self.getkey()
            if s == "KEY_RESIZE":
                self.update()
                return
            self.clear_footer()
            if not (s in ['1', '2', '3', '4', 's', 'e', 'd', 'p']):
                return
            if s in ['1', 's']:
                ed = event.date+datetime.timedelta(seconds=event.duration)
                sd = self.ask_starttime()
                if sd == "CUL_CANCEL":
                    return   # cancel edit event
                if sd > ed:  # shift the event to the new starting hour
                    d = event.duration
                else:  # keep the same end hour
                    # recompute duration
                    d = (ed-sd).seconds
            if s in ['2', 'e']:
                ed = self.ask_endtime(event.date)
                if ed == "CUL_CANCEL":
                    return  # cancel edit event
                d = (ed - event.date).seconds
            if s in ['3', 'd']:
                desc, place = self.ask_description(event)
                if desc == "CUL_CANCEL":
                    return  # canceled edition
            if s in ['4', 'p']:
                place = self.ask_place(event)
                if place == "CUL_CANCEL":
                    return  # canceled edition
            # update the event
            if s in ['1', 's']:
                event.date = sd
            if s in ['1', 's', '2', 'e']:
                event.duration = d
            if s in ['3', 'd']:
                event.summary = desc
                if place:  # non empty
                    event.place = place
            if s in ['4', 'p']:
                event.place = place
            # sort the events
            self._cal.sort()
            # recompute calweek
            self._calweek = calendar.extract_week(self._cal, self._caldav,
                                                  self._day, self,
                                                  autosync=0)
            if event.caldav:  # update the corresponding caldav
                culdav.del_event(event)
                culdav.add_event(event, event.caldav)
            # redraw full day
            self.draw_day()

    def edit_todo(self):
        # the footer will be useful, clear it
        self.clear_footer()
        if len(self._todo.events) < self._todohl-1:
            return
        todo = self._todo.events[self._todohl-1]
        question = _("Edit: (1) [i]tem, (2) [d]ate, (3) [r]emove date")
        self._screen.addstr(self._ymax-2, 0, question)
        s = self.getkey()
        if s == "KEY_RESIZE":
            self.update()
            return
        if not (s in ['1', '2', '3', 'i', 'd', 'r']):
            self.draw_todo()
            return
        if s in ['1', 'i']:
            item, day = self.ask_todo(todo)
            if day:
                todo.date = day
            todo.summary = item
        if s in ['2', 'd']:
            day = self.ask_day()
            if type(day) == datetime.date:
                todo.date = day
        if s in ['3', 'r']:
            todo.date = None
        self._todo.sort()
        self.clear_todo()
        self.draw_todo()

    def add_event(self):
        # the footer will be useful, clear it
        self.clear_footer()

        if self._todohl > 0:
            self.add_todo()
            return

        sd = self.ask_starttime()
        if sd == "CUL_CANCEL":
            return  # cancel add event
        ed = self.ask_endtime(sd)
        if ed == "CUL_CANCEL":
            return  # cancel add event
        duration = (ed - sd).seconds
        desc, place = self.ask_description()
        if desc == "CUL_CANCEL":
            return  # cancel add event

        # create event
        e = calendar.Event(sd, duration, desc, place)
        self._cal.add_event(e)
        self._cal.sort()
        # recompute calweek
        self._calweek = calendar.extract_week(self._cal, self._caldav,
                                              self._day, self,
                                              autosync=0)
        calday = self._calweek[self._day.isoweekday()-1]

        # determine the ehl to set to hl this event
        self._prevehl = self._ehl
        self._ehl = calday.events.index(e)
        # draw new event
        self.draw_day()

    def add_todo(self):
        summ, date = self.ask_todo()
        if summ == "CUL_CANCEL":
            return  # canceled new item
        item = calendar.Event(date, None, summary=summ, tag=0)
        self._todo.add_event(item)
        self._todo.sort()
        self.clear_todo()
        self.draw_todo()

###############################################################################
#       ask questions
###############################################################################

    def ask_generic(self, question, default=""):
        self.clear_footer()
        self._screen.addstr(self._ymax-2, 0, question[:self._width])
        data = inputs.Input(self._screen, default)
        r = data.get_input()
        while r == -1:
            self.update()  # we got a KEY_RESIZE while getting input
            self.clear_footer()
            self._screen.addstr(self._ymax-2, 0, question[:self._width])
            data.screen_update()
            r = data.get_input()

        self.clear_footer()
        if r == 0:
            return data
        else:  # cancel the question
            return "CUL_CANCEL"

    def ask_starttime(self):
        question = _("Enter start time ([hh:mm], [hhmm], [hh] or [h]):")
        start_time = None
        while start_time is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                start_time = data.check_hour(self._day, start=True)
            else:
                data = "CUL_CANCEL"
                start_time = 0
        return self.post_check_input(data, start_time)

    def ask_endtime(self, start_date):
        # ask endtime, return duration of event
        question = (
       _("Enter end time ([hh:mm], [hhmm], [hh] or [h]) or duration ([+mm]):"))
        end_date = None
        while end_date is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                end_date = data.check_duration(self._day, start_date)
            else:
                data = "CUL_CANCEL"
                end_date = 0
        return self.post_check_input(data, end_date)

    def ask_description(self, event=None):
        question = _("Enter description[@location]:")
        if event:
            data = self.ask_generic(question, event.summary)
        else:
            data = self.ask_generic(question)
        try:
            desc = data.text
        except:
            return "CUL_CANCEL", ""  # empty place
        # is there a location in the description?
        if "@" in desc:
            place = desc[desc.find("@")+1:]
            desc = desc[:desc.find("@")]
        else:
            place = ""
        return desc, place

    def ask_place(self, event):
        question = _("Enter location:")
        data = self.ask_generic(question, event.place)
        try:
            return data.text
        except:
            return "CUL_CANCEL"

    def ask_day(self):
        question = (
        _("Enter day ([yyyy/]mm/dd] or [yyyy]mmdd, [Enter for today]):"))
        new_day = None
        while new_day is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                new_day = data.check_day(self._day)
            else:
                data = "CUL_CANCEL"
                new_day = 0
        return self.post_check_input(data, new_day)

    def ask_filename(self, mode, question):
        filename = None
        while filename is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                filename = data.check_filename(mode)
            else:
                data = "CUL_CANCEL"
                filename = 0
        return self.post_check_input(data, filename)

    def tag_event(self):
        if self._todohl > 0:
            return  # we're in the todolist, nothing to do

        self.clear_footer()
        # select highlighted event if existing
        iday = self._day.isoweekday()-1
        if len(self._calweek[iday].events) > self._ehl:
            event = self._calweek[iday].events[self._ehl]
            # the footer will be useful, clear it
            self.clear_footer()
            # ask question
            self._screen.addstr(self._ymax-2, 0,
                                _("Enter tag number [0-7] or [a-z]:"))
            str_tmp = _("(Current tag: ") + str(event.tag) + ")"
            self._screen.addstr(self._ymax-1, 0, str_tmp)
            prekey = self.getkey()
            if prekey == "KEY_RESIZE":
                self.update()
            key = prekey[0]
            if key.isdigit() and (0 <= int(key) <= 7):  # local event
                event.tag = int(key)
                self.clear_footer()
            elif key.isalpha():  # caldav event
                if key in self._caldav._tags:
                    idx = self._caldav._tags.index(key)
                    caldav = self._caldav._icals[idx]

                    if self._caldav._iscaldav[idx]:  # caldav
                        try:  # remove from local events
                            self._cal.del_event(event)
                        except:  # it's a switch from another caldav
                            # not a local event; duplicate
                            e = calendar.Event(event.date, event.duration,
                                               event.summary, event.place,
                                               event.tag, event.url,
                                               event.caldav, event.colour)
                            event = e
                        # switch from local to caldav local culendars
                        event.tag = key
                        event.colour = self._caldav._list[idx]["colour"]
                        self._caldav._cal.add_event(event)
                        # properly add to the extern caldav
                        culdav.add_event(event, caldav)
                        # inform the switch
                        msg = _("Event moved to ")
                        url = caldav.url.url_raw
                        self.inform(msg+url)
                    else:  # it's a webcal, read-only
                        msg = _("Error: can't add an event to a webcal")
                        self._screen.addstr(self._ymax-2, 0, msg)
                        self._screen.addstr(self._ymax-1, 0,
                                            _("Press any key to continue"))
                        key = self.getkey()
                    if key == "KEY_RESIZE":
                        self.update()
                    self.clear_footer()
                else:
                    errmsg = _(" is not an exisiting calDAV tag")
                    self.inform(key+errmsg)
            else:
                self.clear_footer()
                self.inform(_("Incorrect tag"))
            # redraw with new colour
            self._calweek = calendar.extract_week(self._cal, self._caldav,
                                                  self._day, self,
                                                  autosync=0)
            self.draw_day()

    def copy_event(self, repetitions=1):
        if self._todohl > 0:
            return  # we're in the todolist, nothing to do

        iday = self._day.isoweekday()-1
        if len(self._calweek[iday].events) > self._ehl:
            # only if an event is selected
            if repetitions > 1:
                question = _("Copy to next [d]ays, [w]eeks, [m]onths or [y]ears?")
            else:
                question = _("Copy to next [d]ay, [w]eek, [m]onth or [y]ear?")
            self.inform(question)
            key = self.getkey()
            if key == "KEY_RESIZE":
                self.update()
            if key not in ["d", "D", "w", "W", "m", "M", "y", "Y"]:
                errmsg = _("Invalid answer -- canceled")
                self.inform(errmsg)
                return
            event = self._calweek[iday].events[self._ehl]
            date = event.date
            duration = event.duration
            summary = event.summary
            place = event.place
            tag = event.tag
            colour = event.colour
            for ind in range(repetitions):
                if key in ["d", "D"]:
                    new_date = date + datetime.timedelta(days=(ind+1))
                if key in ["w", "W"]:
                    new_date = date + datetime.timedelta(weeks=(ind+1))
                elif key in ["m", "M"]:
                    years = (date.month + ind)//12
                    month = date.month + ind + 1 - 12*years
                    try:
                        new_date = date.replace(year=date.year+years,
                                                month=month)
                    except:
                        new_date = None
                elif key in ["y", "Y"]:
                    new_date = date.replace(year=date.year+(ind+1))
                if new_date:
                    e = calendar.Event(new_date, duration, summary,
                                       place, tag, colour=colour)
                    # deal with tags
                    if type(tag) == int:  # local event
                        self._cal.add_event(e)
                        msg = _("Event copied")
                    elif tag in self._caldav._tags:  # caldav event
                        idx = self._caldav._tags.index(tag)
                        caldav = self._caldav._icals[idx]
                        if self._caldav._iscaldav[idx]:  # caldav
                            self._caldav._cal.add_event(e)
                            # properly add to the extern caldav
                            culdav.add_event(e, caldav)
                            msg = _("Caldav event remotely copied")
                        else:  # webcal
                            self._cal.add_event(e)
                            msg = _("Webcal event locally copied")
            self._cal.sort()
            if key in ["d", "D"]:
                self.draw_week()  # cause the current week changed
            self.inform(msg)
        # no else: do nothing without an selected event

    def shift_event(self, hours):
        if self._todohl > 0:
            return  # we're in the todolist, nothing to do

        # select highlighted event if existing
        iday = self._day.isoweekday()-1
        if len(self._calweek[iday].events) > self._ehl:
            event = self._calweek[iday].events[self._ehl]

            # we can't edit webcal events
            if event.caldav == "webcal":
                # webcals are read-only
                msg = _("Error: can't change webcal event")
                return  # exit here

            sd = event.date+datetime.timedelta(hours=hours)
            event.date = sd

            if event.caldav:  # update the corresponding caldav
                culdav.del_event(event)
                culdav.add_event(event, event.caldav)

            # give focus to moved event

            # set a new day
            self._prevday = self._day
            self._day = sd.date()
            self.avoid_WE(-1)  # new day a non displayed WE: go on Friday

            # update the days of the current week
            self._calweek = calendar.extract_week(self._cal, self._caldav,
                                                  self._day, self,
                                                  self.conf.autosync)

            # search for the event of the new day to highlight the moved one
            iday = self._day.isoweekday()-1
            for i, e in enumerate(self._calweek[iday].events):
                if e.summary == event.summary and e.place == event.place:
                    self._ehl = i
            self.draw_week()

    def ask_todo(self, todo=None):
        question = _("Enter Todo[@date] ([yyyy/]mm/dd] or [yyyy]mmdd):")
        if todo:
            data = self.ask_generic(question, todo.summary)
        else:
            data = self.ask_generic(question)
        if data == "CUL_CANCEL":
            return "CUL_CANCEL"
        try:
            desc = data.text
            day = None
        except:
            return "CUL_CANCEL", None  # None date
        if "@" in desc:  # is there a date?
            date = desc[desc.find("@")+1:]
            item = desc[:desc.find("@")]
            data.text = date
            day = data.check_day(self._day)
        else:
            item = desc
        if not type(day) == datetime.date:
            day = None
        return item, day

    def post_check_input(self, data, processed_data):
        if (data == "KEY_RESIZE") or (processed_data == "KEY_RESIZE"):
            self.update()
            return "CUL_CANCEL"
        if (data == "CUL_CANCEL") or (processed_data == "CUL_CANCEL"):
            self.clear_footer()
            return "CUL_CANCEL"
        else:
            return processed_data

###############################################################################
#       drawing functions
###############################################################################

    def draw_cal(self):
        if self.conf.mouse:  # limit to clickable without bug
            self.draw_mousedeath()
        self.draw_table()
        self.draw_hours()
        self.draw_week()
        if self.conf.todo:
            self.draw_todo()

    def draw_mousedeath(self):
        ymax, real_xmax = self._screen.getmaxyx()
        if self.conf.todo:
            xmax = self._xmax + self._todocol
        else:
            xmax = self._xmax
        if real_xmax > xmax:
            msg = _(" ☠ DON'T CLICK HERE ☠ ")
            nbmsg = (real_xmax-xmax) // len(msg)
            remain = (real_xmax-xmax) - len(msg)*nbmsg
            msg = msg*nbmsg + msg[:remain]
            for y in range(self._ymax-1):
                self._screen.addstr(y, xmax, msg, curses.A_REVERSE)
            self._screen.addstr(self._ymax-1, xmax, msg[:-1], curses.A_REVERSE)

    def draw_table(self):
        # draw vertical lines to bottom point
        bottom = self._lines[1][-1]-2
        self._screen.vline(2, 0, curses.ACS_VLINE, bottom)
        for ld in self._lines[0]:
            self._screen.vline(2, ld, curses.ACS_VLINE, bottom)

        # draw horizontal lines up to the end
        self._screen.hline(1, 1, curses.ACS_HLINE, self._xmax-2)
        for lh in self._lines[1]:
            self._screen.hline(lh, 1, curses.ACS_HLINE, self._xmax-2)

        for lh in self._lines[1]:
            for ld in self._lines[0]:
                # draw crosses
                self._screen.addch(lh, ld, curses.ACS_PLUS)
            # draw left and right tees
            self._screen.addch(lh, 0, curses.ACS_LTEE)
            self._screen.addch(lh, self._lines[0][-1], curses.ACS_RTEE)

        for ld in self._lines[0]:
            # draw top and bottom tees
            self._screen.addch(1, ld, curses.ACS_TTEE)
            self._screen.addch(self._lines[1][-1], ld, curses.ACS_BTEE)

        # draw corners
        self._screen.addch(1, 0, curses.ACS_ULCORNER)
        self._screen.addch(1, self._lines[0][-1], curses.ACS_URCORNER)
        self._screen.addch(self._lines[1][-1], 0, curses.ACS_LLCORNER)
        self._screen.addch(self._lines[1][-1], self._lines[0][-1],
                           curses.ACS_LRCORNER)

    def draw_hours(self):
        self._screen.addstr(2, 1, _('hours'))
        for i, h in enumerate(range(self._hmin, self._hmax)):
            if h < 10:
                self._screen.addstr(self._lines[1][i], 3, '{}h'.format(h))
            else:
                self._screen.addstr(self._lines[1][i], 2, '{}h'.format(h))

    def draw_header(self):
        # erase previous header
        self._screen.hline(0, 0, " ", self._xmax-1)
        if self._todohl:
            style = curses.A_NORMAL
        else:
            style = curses.A_BOLD
        w = (_(' Week ') + str(self._day.isocalendar()[1]) + ', '
             + self._day.strftime("%B") + ' '
             + str(self._day.isocalendar()[0]))
        self._screen.addstr(0, (self._xmax-len(w))//2+1, w, style)
        self._screen.addstr(0, 0, str(self.conf.keys['help'][0]),
                            curses.A_BOLD)
        self._screen.addstr(0, 1, _(": help"), curses.A_BOLD)

    def draw_footer(self, e):
        enddate = e.date + datetime.timedelta(seconds=e.duration)
        self._screen.addstr(self._ymax-3, 0,
                            _('Begin date: {}').format(e.date))
        self._screen.addstr(self._ymax-2, 0,
                            _('End date:   {}').format(enddate))
        if e.place:
            desc = e.summary + " @ " + e.place
        else:
            desc = e.summary
        text = _('Summary: ') + desc
        self._screen.addstr(self._ymax-1, 0, text[:self._width-1])

    def draw_week(self):
        self.draw_header()
        self.draw_daynames()  # redraw day names
        # extraction of the agenda of all interesting days
        self._calweek = calendar.extract_week(self._cal, self._caldav,
                                              self._day, self,
                                              self.conf.autosync)

        if self.conf._debug:
            for iday in range(self._daynb):
                self.draw_day(iday)
            self.draw_hl()
        else:
            try:
                for iday in range(self._daynb):
                    self.draw_day(iday)
                self.draw_hl()
            except:  # couldn't draw it: too small terminal
                self.dead_duck()

    def draw_day(self, iday=-1, hl=0):
        self.clear_day(iday)

        if iday == -1:  # defaults to current day
            iday = self._day.isoweekday()-1
            hl = 1  # not in draw_week, we'll draw the highlight
        calday = self._calweek[iday]

        for e in calday.events:
            self.draw_event(e, curses.A_NORMAL)
        if hl:
            self.draw_hl()

    def draw_hl(self):
        if self.conf.todo and self._todohl > 0:
            self.draw_todo()  # erase previous hl, draw new one
            return
        # else: normal event hl
        self.clear_hl()
        self.draw_daynames()
        iday = self._day.isoweekday()-1
        maxhl = len(self._calweek[iday].events)
        # is there anything to highlight?
        if (self._ehl < maxhl) & (maxhl > 0):
            self.draw_event(self._calweek[iday].events[self._ehl],
                            curses.A_BOLD)
            self.draw_footer(self._calweek[iday].events[self._ehl])

    def draw_daynames(self):
        # first day is monday = 1
        curday = datetime.date.fromordinal(
                    self._day.toordinal() - self._day.isoweekday() + 1)
        for id in range(self._daynb):
            d = curday.strftime("%a %d/%m")
            pos = self._lines[0][id] + (self._day_size - len(d)) // 2 + 1
            if curday == self._day:
                self._screen.addstr(2, pos, d, curses.A_BOLD)
            else:
                self._screen.addstr(2, pos, d)
            # increment day
            curday = datetime.date.fromordinal(curday.toordinal() + 1)

    def draw_event(self, e, flag):
        # flag is bold if highlighted, normal if not
        try:
            colour = self.conf.colours[e.tag] + curses.A_REVERSE + flag
        except:  # alphabetic tag, that's a caldav, use e.color
            colour = self.conf.colours[e.colour] + curses.A_REVERSE + flag

        normal = curses.A_REVERSE + flag

        if type(e.hlines) == tuple:  # cf calendary.py compute_hlines()
            dduck = [_("\_x<  too   >x_/")[0:e.hlines[1]],
                     _("\_x<  many  >x_/")[0:e.hlines[1]],
                     _("\_x< events >x_/")[0:e.hlines[1]]]
            for iv, v in enumerate(e.vlines):
                self._screen.addstr(v, e.hlines[0], dduck[iv % 3])
            return

        width = len(e.hlines)
        # draw background
        for v in e.vlines:
            if self.conf.colourset == 0:
                self._screen.hline(v, e.hlines[0], " ", width, colour)
            elif self.conf.colourset == 1:
                self._screen.addstr(v, e.hlines[0], " ", colour)
                self._screen.addstr(v, e.hlines[0]+width-1, " ", colour)
                self._screen.hline(v, e.hlines[0]+1, " ", width-2, normal)
            elif self.conf.colourset == 2:
                self._screen.hline(v, e.hlines[0], " ", width, normal)
            else:
                self._screen.hline(v, e.hlines[0], " ", width, normal)

        # prepare the text
        t = e.summary
        if len(t) > (width):
            t = t[0:width]

        # draw text for all cases, bold for highlighted event
        posx = e.hlines[0] + (width - len(t))//2
        posy = e.vlines[0] + (e.vlines[-1]-e.vlines[0])//2
        if self.conf.colourset == 0:
            self._screen.addstr(posy, posx, t, colour)
        elif self.conf.colourset == 1:
            self._screen.addstr(posy, posx, t, normal)
            if len(t) == width:
                self._screen.addstr(posy, posx, t[0], colour)
                self._screen.addstr(posy, posx+width-1, t[-1], colour)
        elif self.conf.colourset == 2:
            self._screen.addstr(posy, posx, t, colour)
        else:
            self._screen.addstr(posy, posx, t, normal)

        # if there is a location, add it if it's a two lines event
        if (e.place != "") & (e.vlines[-1] > e.vlines[0]):
            # prepare the text
            p = e.place
            if len(p) > (width):
                p = p[0:width]
            posx = e.hlines[0] + (width - len(p))//2
            if self.conf.colourset == 0:
                self._screen.addstr(posy+1, posx, p, colour)
            elif self.conf.colourset == 1:
                self._screen.addstr(posy+1, posx, p, normal)
                if len(p) == width:
                    self._screen.addstr(posy+1, posx, p[0], colour)
                    self._screen.addstr(posy+1, posx+width-1, p[-1], colour)
            elif self.conf.colourset == 2:
                self._screen.addstr(posy+1, posx, p, colour)
            else:
                self._screen.addstr(posy+1, posx, p, normal)

    def draw_todo(self):
        if self._todohl:
            style = curses.A_BOLD
        else:
            style = curses.A_NORMAL
        title = _("TODO")
        xpos = self._xmax + (self._todocol-len(title))//2
        self._screen.addch(0, self._xmax-1, curses.ACS_VLINE)
        self._screen.addch(1, self._xmax-1, curses.ACS_PLUS)
        self._screen.addch(self._lines[1][-1], self._xmax-1, curses.ACS_RTEE)
        self._screen.vline(self._lines[1][-1]+1, self._xmax-1,
                           curses.ACS_VLINE, self._ymax-self._lines[1][-1]-4)
        self._screen.hline(1, self._xmax, curses.ACS_HLINE, self._todocol)
        self._screen.addstr(0, xpos, title, style)

        for idx, item in enumerate(self._todo.events):
            if idx+1 == self._todohl:
                style = curses.A_BOLD
            else:
                style = curses.A_NORMAL
            if item.date:
                date = item.date.strftime("%d/%m")
                self._screen.addstr(2+idx, self._xmax+self._todocol-len(date),
                                    date, style)
            if len(item.summary) < self._todocol:
                # + " " to ensure a space before the date
                text = item.summary + " "
            else:
                text = item.summary[:self._todocol-1]
            self._screen.addstr(2+idx, self._xmax+1, text, style)
        if self._todohl > 0 and len(self._todo.events) > 0:
            self.draw_todohl()

    def draw_todohl(self):
        # draw highlighted todo in the footer
        self.clear_footer()
        if len(self._todo.events)+1 > self._todohl:
            item = self._todo.events[self._todohl-1]
            date = ""
            if item.date:
                date = (" -- " + str(item.date.day) + "/"
                               + str(item.date.month) + "/"
                               + str(item.date.year))
            text = item.summary + date
            self._screen.addstr(self._ymax-1, 0, text[:self._width-1])

###############################################################################
#       clearing functions
###############################################################################

    def clear_cal(self):
        self._screen.clear()

    def clear_week(self):
        self.clear_footer()
        # clear the previous week
        prevcalweek = calendar.extract_week(self._cal, self._caldav,
                                            self._prevday, self,
                                            autosync=0)
        for A in prevcalweek[0:self._daynb]:
            for e in A.events:
                self.clear_event(e)

    def clear_day(self, iday=-1):
        if iday == -1:  # defaults to current day
            iday = self._day.isoweekday()-1

        # if last iday, the day size is not self._day_size
        if iday == self.daynb-1:
            width = self.lines[0][-1] - self.lines[0][-2] - 1
        else:
            width = self.day_size - 1
        # day_size - 1: doesn't include vertical separating lines

        for v in range(self.lines[1][0], self.lines[1][-1]+1):
            # +1 to be sure to include last line
            self._screen.hline(v, self.lines[0][iday]+1, " ", width)
            if v in self.lines[1]:
                # if it's a horizontal line, redraw it
                self._screen.hline(v, self.lines[0][iday]+1,
                                   curses.ACS_HLINE, width)

    def clear_event(self, e):
        # draw spaces to erase the background
        width = len(e.hlines)
        for v in e.vlines:
            self._screen.hline(v, e.hlines[0], " ", width)
            if v in self.lines[1]:
                # if it's a horizontal line, redraw it
                self._screen.hline(v, e.hlines[0],
                                   curses.ACS_HLINE, width)

    def clear_hl(self):
        self.clear_footer()
        iday = self._day.isoweekday()-1
        previday = self._prevday.isoweekday()-1
        # we toggled the WE off; nothing to do
        if previday >= self._daynb:
            pass
        # if we change the hl of current day
        elif self._prevday == self._day:
            # just redraw in normal font
            iday = self._day.isoweekday()-1
            maxhl = len(self._calweek[iday].events)
            # is there anything to highlight?
            if (self._prevehl < maxhl) & (maxhl > 0):
                self.draw_event(self._calweek[iday].events[self._prevehl],
                                curses.A_NORMAL)
        # if we change from less than a week
        # same isoweek and same year
        elif self._day.isocalendar()[:2] == self._prevday.isocalendar()[:2]:
            # redraw previous day
            calday = self._calweek[self._prevday.isoweekday()-1]
            for e in calday.events:
                self.draw_event(e, curses.A_NORMAL)
        # else: nothing to do, new week clears everything

    def clear_todo(self):
        for y in range(2, self._ymax):
            self._screen.hline(y, self._xmax, " ", self._todocol)

    def clear_footer(self):
        for y in range(self._ymax-3, self._ymax):
            self._screen.hline(y, 0, " ", self._width)

###############################################################################
# save/import/export
###############################################################################

    def save(self):
        calendar.save(self._cal, self._todo, self._conf, self._caldav)

    def importcal(self):
        question = (
       _("Enter the file name to import (ical or calcurse): (filename [tag])"))
        filename = self.ask_filename("rb", question)
        tag = 0  # default value
        if filename == "CUL_CANCEL":
            return
        if ((filename[-2] == " ")
                and (filename[-1] in [str(i) for i in range(8)])):
            tag = int(filename[-1])
            filename = filename[:-2]
        try:  # is it an ical?
            cal = calendar.import_ical(filename, self.conf.categories, tag)
        except:
            try:  # is it from calcurse?
                cal = calendar.import_calcurse(filename, tag)
            except:
                self._screen.addstr(self._ymax-2, 0, _("Unable to load file"))
                self._screen.addstr(self._ymax-1, 0,
                                    _("Press any key to continue"))
                key = self.getkey()
                if key == "KEY_RESIZE":
                    self.update()
                self.clear_footer()
                return

        # check for already present event
        # 1) create simple list from self._cal
        existingevents = []
        for se in self._cal.events:
            existingevents.append(se.date.strftime('%Y%m%d%H%M')
                                  + str(se.duration)
                                  + se.summary.replace(' ', ''))
        for e in cal.events:
            currentevent = (e.date.strftime('%Y%m%d%H%M')
                            + str(e.duration)
                            + e.summary.replace(' ', ''))
            if not(currentevent in existingevents):
                self._cal.add_event(e)
        self._cal.sort()
        self.draw_cal()
        self.inform(_("New events imported!"))

    def exportcal(self):
        # returns 1 if successful
        self.clear_footer()
        key = ""
        while not(key in ["i", "I", "c", "C", "q", "Q", ""]):
            self._screen.addstr(self._ymax-2, 0,
                                _("Export to [i]cal or to [c]alcurse?"))
            self._screen.addstr(self._ymax-1, 0, "[i/c]")
            key = self.getkey()
            if key == "KEY_RESIZE":
                self.update()
                return
            if key in ["q", "Q", ""]:
                self.clear_footer()
                return
        question = _("Enter the file name to export:")
        filename = self.ask_filename("wb", question)
        if filename == "CUL_CANCEL":
            return

        if key in ["c", "C"]:
            calendar.export_calcurse(self._cal, filename)
        else:
            calendar.export_ical(self._cal, filename, self.conf.categories)
        self.inform(_("Calendar exported!"))

###############################################################################
# other stuff (waiting for a better section name)
###############################################################################

    def getkey(self):
        return getkey.getkey(self._screen, self.conf._debug)

    # use footer to give feedback
    def inform(self, msg):
        self.clear_footer()
        self._screen.addstr(self._ymax-1, 0, msg[0:self._width-1])
        self._screen.refresh()

    def debug(self, elt):
        self._screen.addstr(self._ymax-2, 50,
                            'debug:   {}'.format(elt))
        self._screen.getch()  # just a pause
        self._screen.hline(self._ymax-2, 1, " ", self._width-2)

    def sync_caldav(self):
        # update caldavs only
        self._caldav.sync()
        # redraw if something changed
        self.clear_cal()
        self.draw_cal()



#! /usr/bin/env python3
# coding: utf-8
import curses
from . import inputs, getkey
from collections import OrderedDict


class Conf:
    def __init__(self, scr, conf, caldav):
        self._screen = scr
        self._caldav = caldav
        self._y, self._x = self._screen.getmaxyx()
        self._top = 0  # offset between screen and pad
        self._conf = conf
        self._general = OrderedDict()
        self._general["hmin"] = conf.hmin
        self._general["hmax"] = conf.hmax
        self._general["WE"] = conf.WE
        self._general["TODO"] = conf.todo
        self._general["autosave"] = conf.autosave
        self._general["colourset"] = conf.colourset
        self._general["autosync"] = conf.autosync
        self._general["mouse"] = conf.mouse
        self._tab = 0
        self._tabs = [self._general,
                      self._conf.keys,
                      self._conf.categories,
                      self._conf.caldav]
        self._item = 0
        self._tabsnames = [_("General configuration"),
                           _("Keybinding"),
                           _("Categories"),
                           _("CalDAV")]
        # to be independant from config.set_colours
        curses.init_pair(42, curses.COLOR_RED, -1)
        curses.init_pair(43, curses.COLOR_GREEN, -1)

########################################################################
# drawing functions
########################################################################

    def draw_conf_screen(self):
        self._screen.clear()
        self._y, self._x = self._screen.getmaxyx()
        self._caldav.resize()  # update xmax and ymax

        # if ridiculously small terminal
        if (self._x < 25) or (self._y < 6):
            self._screen.clear()
            text = _("\_x< A dead duck. That's all such a small terminal deserves.")
            self._screen.addstr(0, 0, text[0:(self._x-1)*(self._y-1)])

            key = ""
            while not((key in self._conf.keys["quit"])
                      or (key == "KEY_RESIZE")):
                key = getkey.getkey(self._screen, self._conf.debug)
            if key in self._conf.keys["quit"]:
                exit()
            else:
                self.draw_conf_screen()

        # draw screen
        headleft = _("q: quit configuration")
        headcenter = _("Culendar configuration screen")
        self._screen.border()
        self._screen.hline(2, 1, curses.ACS_HLINE, self._x-2)
        self._screen.addch(2, 0, curses.ACS_LTEE)
        self._screen.addch(2, self._x-1, curses.ACS_RTEE)
        self._screen.addstr(1, 1, headleft, curses.A_BOLD)
        self._screen.addstr(1, (self._x-len(headcenter))//2+1,
                            headcenter, curses.A_BOLD)

        # should be factorized with helpscr.py lines 83+
        # create full pad in case of terminal smaller than needed
        # counting lines for each tab, take max + 2 lines for tabnames
        self._padlines = 2 + max([len(d) for d in self._tabs])
        # pad columns: very hard to set a limit, one could add as many
        # keybinding as he wants… Set to 10×sizeX, to have lots of room
        # even for 80col terminals
        self._padcols = 10*self._x
        self._pad = curses.newpad(self._padlines, self._padcols)
        self._pad.scrollok(True)
        self._screen.refresh()  # if not present, pad is ignored

        # the lines available for the pad
        self._padavail = self._y - 4
        # to have the bottom of the pad on the bottom of the screen
        self._maxtop = max(0, self._padlines - self._padavail)

        self.draw_tabs()

    def draw_tabs(self):
        # draw tabs
        colsize = round((self._x - 2) / len(self._tabs))
        self._pad.hline(1, 0, curses.ACS_HLINE, self._x-1)

        for it, tname in enumerate(self._tabsnames):
            if it > 0:  # add previous separator
                self._pad.addch(0, it*colsize, curses.ACS_VLINE)
                self._pad.addch(1, it*colsize, curses.ACS_BTEE)
            if it == self._tab:
                OPT = curses.A_BOLD
            else:
                OPT = curses.A_NORMAL
            tabname = tname[0:colsize]  # to be sure to fit
            start = round((it + 1/2)*colsize - len(tabname)/2)
            self._pad.addstr(0, start, tabname, OPT)

        if self._tab < 3:
            self.draw_dictionary()
        elif tname == "CalDAV":
            self.draw_caldav()
        self.draw_help()
        self._pad.refresh(self._top, 0, 3, 1, self._y-2, self._x-2)

    def draw_dictionary(self):
        if self._tab != 0:  # General is already an OrderDict
            stuff = sorted(self._tabs[self._tab].keys())
        else:
            stuff = self._tabs[self._tab].keys()

        for ik, k in enumerate(stuff):
            # key k is integer for some dict, gets str(.) when useful
            if ik == self._item:
                OPT = curses.A_BOLD
                # save the key to use it in edit_item()
                self._editing_key = k
            else:
                OPT = curses.A_NORMAL
            # if k is integer, it's the category number, rename it
            if type(k) == int:
                # add the associated color
                COL_OPT = self._conf.colours[k] + curses.A_REVERSE+OPT
                printme = _("Category ") + str(k)
                self._pad.addstr(2+ik, 2, printme, COL_OPT)
                offset = len(printme)
            else:
                self._pad.addstr(2+ik, 2, k, OPT)
                offset = len(k)
            if self._editing_key == k:
                # remember this offset for deleting item
                self._offset = offset

            self._pad.addstr(2+ik, 2+offset, " = ", OPT)
            if type(self._tabs[self._tab][k]) == bool:
                if self._tabs[self._tab][k]:
                    self._pad.addstr(2+ik, 2+offset+3, _("On"),
                                     OPT+curses.color_pair(43))
                else:
                    self._pad.addstr(2+ik, 2+offset+3, _("Off"),
                                     OPT+curses.color_pair(42))
            elif self._tab != 0:  # editing keys or categories
                listkeys = ""
                for ck in self._tabs[self._tab][k]:
                    # special case: keys (cf helpscr.py)
                    if ck == "\t":
                        listkeys += "TAB, "
                    elif ck == " ":
                        listkeys += "SPACE, "
                    else:
                        listkeys += ck+", "
                listkeys = listkeys[:-2]  # erase final comma
                self._pad.addstr(2+ik, 2+offset+3, listkeys, OPT)
            else:
                self._pad.addstr(2+ik, 2+offset+3,
                                 str(self._tabs[self._tab][k]), OPT)
            if k == "colourset":  # special case, add an example of colourset
                example = _("Example")
                width = len(example) + 6
                colour = curses.A_REVERSE + curses.color_pair(42) + OPT
                normal = curses.A_REVERSE + OPT
                if self._tabs[self._tab][k] == 0:
                    self._pad.hline(2+ik, 2+offset+3+3, " ",
                                    width, colour)
                    self._pad.addstr(2+ik, 2+offset+3+6,
                                     example, colour)
                elif self._tabs[self._tab][k] == 1:
                    self._pad.addstr(2+ik, 2+offset+3+3, " ",
                                     colour)
                    self._pad.addstr(2+ik, 2+offset+3+3+width-1,
                                     " ", colour)
                    self._pad.hline(2+ik, 2+offset+3+3+1, " ",
                                    width-2, normal)
                    self._pad.addstr(2+ik, 2+offset+3+6, example,
                                     normal)
                elif self._tabs[self._tab][k] == 2:
                    self._pad.hline(2+ik, 2+offset+3+3, " ",
                                    width, normal)
                    self._pad.addstr(2+ik, 2+offset+3+6,
                                     example, colour)
                else:
                    self._pad.hline(2+ik, 2+offset+3+3, " ",
                                    width, normal)
                    self._pad.addstr(2+ik, 2+offset+3+6, example,
                                     normal)

    def draw_caldav(self):
        tagcol = _("T")
        urlcol = _("URL")
        usernamecol = _("Username")
        passwordcol = _("Password")
        maxurl = len(urlcol)
        maxusr = len(usernamecol)
        for ik, cd in enumerate(self._conf.caldav):
            maxurl = max(maxurl, len(cd["url"]))
            maxusr = max(maxusr, len(cd["username"]))

        reflen = min(maxurl+maxusr+9, self._x-3)

        # y-position of the columns
        maxlen = self._x-3
        xtag = 2
        xurl = xtag + 2
        if maxurl+maxusr+11 > maxlen:
            # +11: to have place for colour and at least 4 first pwd characters
            # 4 for a minimal characters in pwd, 3 for space, split the rest
            halflen = (maxlen-xurl-4-3) // 2
            # check if url or usr is smaller than this half length and optim
            if maxurl < halflen:
                maxusr = min(maxusr, maxlen - xurl - 7 - maxurl)
            elif maxusr < halflen:
                maxurl = min(maxurl, maxlen - xurl - 7 - maxusr)
            else:
                maxurl = maxusr = halflen
        xusr = xurl + 2 + maxurl
        xpwd = xusr + 2 + maxusr

        # print columns title
        self._pad.addstr(2, xtag, tagcol, curses.A_BOLD)
        self._pad.addstr(2, xurl, urlcol[:maxurl], curses.A_BOLD)
        self._pad.addstr(2, xusr, usernamecol[:maxusr], curses.A_BOLD)
        self._pad.addstr(2, xpwd, passwordcol, curses.A_BOLD)

        for ik, cd in enumerate(self._conf.caldav):
            if ik == self._item:
                OPT = curses.A_BOLD
                # save the key to use it in edit_item()
                self._editing_key = ik
            else:
                OPT = curses.A_NORMAL
            COL_OPT = self._conf.colours[cd["colour"]] + curses.A_REVERSE+OPT
            self._pad.addstr(4+ik, xtag, cd["tag"], COL_OPT)
            if "@" in cd["url"]:  # there is a password to hide in a webcal
                head_user_password, tail = cd["url"].split("@")
                head, user_password = head_user_password.split("://")
                user = user_password.split(":")[0]
                # if there are ":" in the password, join them again
                password = ":".join(user_password.split(":")[1:])
                url = head+"://"+user+":"+"*****"+"@"+tail
            else:
                url = cd["url"]
            self._pad.addstr(4+ik, xurl, url[:maxurl], OPT)
            if cd["url"][:4] == "http":
                self._pad.addstr(4+ik, xusr, cd["username"][:maxusr], OPT)
                self._pad.addstr(4+ik, xpwd, "*"*len(cd["password"]), OPT)

    def draw_help(self):
        self.oneline_footer()
        if self._tab == 0:  # general
            helpmsg = _("Enter/Space: edit value")
        elif self._tab == 1:  # keys
            helpmsg = _("a: add key    d: delete key")
        elif self._tab == 2:  # categories
            helpmsg = _("a: add name   d: delete name")
            helpmsg += _("    Enter/Space: change colour")
        elif self._tab == 3:  # caldav
            helpmsg = _("a: add caldav    d: delete caldav")
            helpmsg += _("    t: change tag    c: change colour")
            helpmsg += _("    u: edit url    n: edit username")
            helpmsg += _("    p: edit password")
        self._screen.addstr(self._y-1, 0, helpmsg[:self._x-1])

    def clear_tabs(self):
        self._pad.hline(0, 1, " ", self._padcols)
        self._pad.clrtobot()

    def redraw_tabs(self):
        self.clear_tabs()
        self.draw_tabs()

    def oneline_footer(self):
        self._screen.addch(self._y-4, self._x-1, curses.ACS_VLINE)
        self._screen.clrtobot()
        self._screen.addch(self._y-3, 0, curses.ACS_VLINE)
        self._screen.addch(self._y-3, self._x-1, curses.ACS_VLINE)
        self._screen.hline(self._y-2, 1, curses.ACS_HLINE, self._x-2)
        self._screen.addch(self._y-2, 0, curses.ACS_LLCORNER)
        self._screen.addch(self._y-2, self._x-1, curses.ACS_LRCORNER)

    def question_footer(self):
        self._screen.hline(self._y-3, 1, curses.ACS_HLINE, self._x-2)
        self._screen.addch(self._y-3, 0, curses.ACS_LLCORNER)
        self._screen.addch(self._y-3, self._x-1, curses.ACS_LRCORNER)
        self._screen.clrtobot()

    def no_footer(self):
        self._screen.addch(self._y-4, self._x-1, " ")
        self._screen.clrtobot()
        self._screen.box()
        self._screen.refresh()

########################################################################
# editing functions
########################################################################

    def edit_item(self):
        if self._tab == 0:  # general
            if type(self._tabs[self._tab][self._editing_key]) == bool:
                self.edit_bool()
            elif self._editing_key == "colourset":
                # very specific case, toggle in different coloursets
                self.edit_colourset()
            elif type(self._tabs[self._tab][self._editing_key]) == int:
                self.edit_int_hour()
        elif self._tab == 2:  # change category colour
            self.edit_change_colour()

    def edit_bool(self):
        self._tabs[self._tab][self._editing_key] = not(
                        self._tabs[self._tab][self._editing_key])

    def edit_colourset(self):
        self._tabs[self._tab][self._editing_key] = (
                        self._tabs[self._tab][self._editing_key] + 1) % 3
        # 3 coloursets avalaible

    def edit_generic(self, question, default="", pwd=False):
        self.question_footer()
        self._screen.addstr(self._y-2, 0, question[:self._x])
        data = inputs.Input(self._screen, default, pwd)
        r = data.get_input()
        while r == -1:
            self.draw_conf_screen()  # we got a KEY_RESIZE while getting input
            self.question_footer()
            self._screen.addstr(self._y-2, 0, question[:self._x])
            data.screen_update()
            r = data.get_input()
        if r == 0:
            return data
        else:  # cancel the question
            return "CUL_CANCEL"

    def edit_int_hour(self):
        question = _("Enter new value [hh] or [h]:")
        new_int = None
        while new_int is None:
            data = self.edit_generic(question)
            if data == "CUL_CANCEL":
                return
            new_int = data.check_inthour()
            if self.post_edit(new_int) == "CUL_CANCEL":
                return
        self._tabs[self._tab][self._editing_key] = new_int

    def edit_delete(self):
        if self._tab == 3:
            # pop the current caldav out of the list if any
            if len(self._tabs[self._tab]) != 0:
                self._tabs[self._tab].pop(self._editing_key)
            if self._editing_key == len(self._tabs[self._tab]):
                # last position
                self._item = len(self._tabs[self._tab]) - 1
        else:
            self.edit_delete_item()     # choose an item to delete

    def edit_delete_item(self):
        if len(self._tabs[self._tab][self._editing_key]) == 0:
            return
        elif len(self._tabs[self._tab][self._editing_key]) == 1:
            self._tabs[self._tab][self._editing_key].pop()
        else:
            self._item_idx = 0
            self.oneline_footer()
            helpmsg = _("Select item to delete, press Enter")
            self._screen.addstr(self._y-1, 0, helpmsg[:self._x-1])
            self.edit_delete_item_draw()
            key = getkey.getkey(self._screen, self._conf.debug)
            while not(key in ["q", ""] + self._conf.keys["quit"]):
                if key == "KEY_RESIZE":
                    self.draw_conf_screen()
                    return
                elif key in (["\t", "KEY_RIGHT"]
                             + self._conf.keys["nextday"]):
                    self._item_idx += 1
                    if self._item_idx > len(
                           self._tabs[self._tab][self._editing_key])-1:
                        self._item_idx = 0
                elif key in (["KEY_BTAB", "KEY_LEFT"]
                             + self._conf.keys["prevday"]):
                    self._item_idx -= 1
                    if self._item_idx < 0:
                        self._item_idx = len(
                            self._tabs[self._tab][self._editing_key])-1
                elif key in ["d", "D", "\n"]:
                    self._tabs[self._tab][self._editing_key].pop(
                                                        self._item_idx)
                    return
                self.edit_delete_item_draw()
                key = getkey.getkey(self._screen, self._conf.debug)

    def edit_delete_item_draw(self):
        msg = ""
        item_pos = [len(msg)]
        # positioning the cursor just after the = sign
        self._pad.addstr(2+self._item-self._top, 2+self._offset+2, " ")
        for i, item in enumerate(
                       self._tabs[self._tab][self._editing_key]):
            if item == "\t":
                item = "TAB"
            elif item == " ":
                item = "SPACE"
            if i == self._item_idx:
                self._pad.addstr(item, curses.A_BOLD)
            else:
                self._pad.addstr(item)
            if i < len(self._tabs[self._tab][self._editing_key])-1:
                self._pad.addstr(", ")
        # may crash if very long categories / keys
        self._pad.refresh(self._top, 0, 3, 1, self._y-2, self._x-2)

    def edit_add_item(self):
        if self._tab == 1:
            self.edit_add_key()
        elif self._tab == 2:
            self.edit_add_category()
        elif self._tab == 3:
            self.edit_add_caldav()

    def edit_add_key(self):
        self.oneline_footer()
        question = _("Press the key to add (Escape to cancel)")
        self._screen.addstr(self._y-1, 0, question[:self._x-1])
        key = getkey.getkey(self._screen, self._conf.debug)
        if key == "KEY_RESIZE":
            self.draw_conf_screen()
        elif key != "":
            if not(key in self._tabs[self._tab][self._editing_key]):
                # nothing to add, already here
                self._tabs[self._tab][self._editing_key].append(key)
            # check if added key is present elsewhere
            for k in self._tabs[self._tab].keys():
                # we don't care about the current edited key
                if k != self._editing_key:
                    if key in self._tabs[self._tab][k]:
                        # the new key was already in use: remove it
                        self._tabs[self._tab][k].pop(
                            self._tabs[self._tab][k].index(key)
                                                    )
                        self.question_footer()
                        warning = (_('Warning: key "')
                                   + str(key)
                                   + _('" was used for function "')
                                   + str(k) + _('"'))
                        wait = _("Press any key to continue")
                        self._screen.addstr(self._y-2, 0,
                                            warning[:self._x-1])
                        self._screen.addstr(self._y-1, 0,
                                            wait[:self._x-1])
                        # pause; resize if necessary
                        if getkey.getkey(self._screen) == "KEY_RESIZE":
                            self.draw_conf_screen()

    def edit_add_category(self):
        question = _("Enter new category name:")
        new_cat = None
        while new_cat is None:
            data = self.edit_generic(question)
            if data == "CUL_CANCEL":
                return
            else:
                new_cat = data.text
        self._tabs[self._tab][self._editing_key].append(new_cat)

    def edit_add_caldav(self):
        new_url = self.enter_url()
        if not new_url:
            return

        question = _("Enter tag [a-z]:")
        self.question_footer()
        self._screen.addstr(self._y-2, 0, question[:self._x])
        new_tag = getkey.getkey(self._screen, self._conf.debug)
        if not new_tag.isalpha():
            self.draw_conf_screen()  # if it is KEY_RESIZE, to be sure
            return
        new_colour = 0  # default colour

        if new_url[:4] == "http":  # caldav: require user/pwd
            new_user = self.enter_username()
            if not new_user:
                return
            new_pwd = self.enter_password()
            if not new_pwd:
                return
        else:
            new_user = ""
            new_pwd = ""
        # create the final dictionary
        cdav = {}
        cdav["url"] = new_url
        cdav["username"] = new_user
        cdav["password"] = new_pwd
        cdav["tag"] = new_tag
        cdav["colour"] = new_colour
        self._conf.caldav.append(cdav)
        self._item = len(self._conf.caldav) - 1

    def enter_url(self, url=""):
        question1 = _("Enter URL: (http(s)://[…] for a Caldav, ")
        question2 = _("webcal(s)://[user:password@][…] for a Webcal)")
        question = question1 + question2
        new_url = None
        while new_url is None:
            data = self.edit_generic(question, url)
            if data == "CUL_CANCEL":
                return
            else:
                new_url = data.text
        if new_url[:4] not in ("http", "webc"):
            err = _("Error: URL starts with http(s):// or webcal://")
            self._screen.addstr(self._y-2, 0, " "*self._x)
            self._screen.addstr(self._y-2, 0, err)
            self._screen.addstr(self._y-1, 0, _("Press any key to continue"))
            key = getkey.getkey(self._screen, self._conf.debug)
            if key == "KEY_RESIZE":
                self.draw_conf_screen()
            return
        return new_url

    def enter_username(self, user=""):
        question = _("Enter username:")
        new_user = None
        while new_user is None:
            data = self.edit_generic(question, user)
            if data == "CUL_CANCEL":
                return
            else:
                new_user = data.text
        return new_user

    def enter_password(self, pwd=""):
        question = _("Enter password:")
        new_password = None
        while new_password is None:
            data = self.edit_generic(question, pwd, pwd=True)
            if data == "CUL_CANCEL":
                return
            else:
                new_password = data.text
        return new_password

    def edit_caldav_tag(self):
        cdav = self.cdav_copy(self._conf.caldav[self._editing_key])
        cur_tag = self._conf.caldav[self._editing_key]["tag"]
        question = _("Enter new tag [a-z]:")
        answer = _("(Current tag: ") + str(cur_tag) + ")"
        self.question_footer()
        self._screen.addstr(self._y-2, 0, question[:self._x])
        self._screen.addstr(self._y-1, 0, answer[:self._x])
        new_tag = getkey.getkey(self._screen, self._conf.debug)
        if not new_tag.isalpha():
            self.draw_conf_screen()  # if it is KEY_RESIZE, to be sure
            return
        cdav["tag"] = new_tag
        self._conf.caldav.pop(self._editing_key)
        self._conf.caldav.insert(self._editing_key, cdav)

    def edit_caldav_colour(self):
        cdav = self.cdav_copy(self._conf.caldav[self._editing_key])
        colour = (self._conf.caldav[self._editing_key]["colour"] + 1) % 8
        cdav["colour"] = colour
        self._conf.caldav.pop(self._editing_key)
        self._conf.caldav.insert(self._editing_key, cdav)

    def edit_caldav_url(self):
        cdav = self.cdav_copy(self._conf.caldav[self._editing_key])
        new_url = self.enter_url(cdav["url"])
        if not new_url:
            return
        cdav["url"] = new_url
        self._conf.caldav.pop(self._editing_key)
        self._conf.caldav.insert(self._editing_key, cdav)

    def edit_caldav_username(self):
        if self.iscaldav():
            cdav = self.cdav_copy(self._conf.caldav[self._editing_key])
            new_user = self.enter_username(cdav["username"])
            if not new_user:
                return
            cdav["username"] = new_user
            self._conf.caldav.pop(self._editing_key)
            self._conf.caldav.insert(self._editing_key, cdav)

    def edit_caldav_password(self):
        if self.iscaldav():
            cdav = self.cdav_copy(self._conf.caldav[self._editing_key])
            new_pwd = self.enter_password(cdav["password"])
            if not new_pwd:
                return
            cdav["password"] = new_pwd
            self._conf.caldav.pop(self._editing_key)
            self._conf.caldav.insert(self._editing_key, cdav)

    def edit_change_colour(self):
        indexes = self._conf.colours_idx
        indexes[self._editing_key] = (indexes[self._editing_key] + 1) % 8
        colours = self._conf.colours
        colours[self._editing_key] = curses.color_pair(
                                            indexes[self._editing_key])
        self._conf.colours = colours

    def post_edit(self, data):
        if (data == "KEY_RESIZE") or (data == "CUL_CANCEL"):
            # during error, resized or did nothing
            self.draw_conf_screen()
            return "CUL_CANCEL"
        else:
            # no error during input
            return data

########################################################################
# other functions
########################################################################

    def localconf_to_culconf(self):
        # due to the setter, if newhmin > oldhmax
        # hmin = oldhmax - 1
        self._conf.hmin = self._general["hmin"]
        self._conf.hmax = self._general["hmax"]
        # with the new hmax, use the setter for new hmin
        self._conf.hmin = self._general["hmin"]
        self._conf.WE = self._general["WE"]
        self._conf.todo = self._general["TODO"]
        self._conf.autosave = self._general["autosave"]
        self._conf.colourset = self._general["colourset"]
        self._conf.autosync = self._general["autosync"]
        self._conf.mouse = self._general["mouse"]

    def conf_screen(self):
        self.draw_conf_screen()
        key = getkey.getkey(self._screen, self._conf.debug)
        while True:
            if key in ["KEY_RESIZE"] + self._conf.keys["redraw"]:
                self.draw_conf_screen()

            if (key in ["a", "A"]) and (self._tab > 0):  # key category caldav
                self.edit_add_item()

            if (key in ["d", "D"]) and (self._tab > 0):  # key category caldav
                self.edit_delete()

            if (key in ["t", "T"]) and (self._tab > 0):  # caldav tag
                if len(self._conf.caldav) > 0:  # at least a caldav to edit
                    self.edit_caldav_tag()

            if (key in ["c", "C"]) and (self._tab > 0):  # caldav colour
                if len(self._conf.caldav) > 0:  # at least a caldav to edit
                    self.edit_caldav_colour()

            if (key in ["u", "U"]) and (self._tab > 0):  # caldav url
                if len(self._conf.caldav) > 0:  # at least a caldav to edit
                    self.edit_caldav_url()

            if (key in ["n", "N"]) and (self._tab > 0):  # caldav username
                if len(self._conf.caldav) > 0:  # at least a caldav to edit
                    self.edit_caldav_username()

            if (key in ["p", "P"]) and (self._tab > 0):  # caldav password
                if len(self._conf.caldav) > 0:  # at least a caldav to edit
                    self.edit_caldav_password()

            if key in [" ", "\n"] + self._conf.keys["editevent"]:
                self.edit_item()

            if key in ["KEY_UP"] + self._conf.keys["prevweek"]:
                if self._item > 0:
                    self._item -= 1
                    if self._item < self._top:
                        self._top -= 1
                else:
                    self._item = len(self._tabs[self._tab])-1
                    self._top = max(0, self._item - (self._y-9))

            if key in ["KEY_DOWN"] + self._conf.keys["nextweek"]:
                if self._item < len(self._tabs[self._tab])-1:
                    self._item += 1
                    # Are we on last on the last pad line?
                    if self._item - self._top == self._y-8:
                        # scroll it
                        self._top += 1
                else:
                    self._item = 0
                    self._top = 0

            if key in ["\t", "KEY_RIGHT"] + self._conf.keys["nextday"]:
                self._tab = (self._tab + 1) % len(self._tabs)
                self._top = 0
                self._item = 0

            if key in ["KEY_BTAB", "KEY_LEFT"] + self._conf.keys["prevday"]:
                self._tab = (self._tab - 1) % len(self._tabs)
                self._top = 0
                self._item = 0

            if key in ["q"] + self._conf.keys["quit"]:
                self.localconf_to_culconf()
                self.oneline_footer()  # in case of message of caldav sync
                return self._conf

            self.redraw_tabs()
            key = getkey.getkey(self._screen, self._conf.debug)

    def iscaldav(self):
        if self._conf.caldav[self._editing_key]["url"][:6] == "webcal":
            self.oneline_footer()
            msg = _("Webcals do not require this parameter")
            self._screen.addstr(self._y-1, 0, msg[:self.x])
            return False
        return True

    def debug(self, elt):
        self._screen.addstr(self._y-2, 50,
                            'debug:   {}'.format(elt))
        self._screen.getch()  # just a pause
        self._screen.hline(self._y-2, 50, " ", self._x-51)

    def cdav_copy(self, cdav):
        new_cdav = {}
        new_cdav["url"] = cdav["url"]
        new_cdav["username"] = cdav["username"]
        new_cdav["password"] = cdav["password"]
        new_cdav["tag"] = cdav["tag"]
        new_cdav["colour"] = cdav["colour"]
        return new_cdav



#! /usr/bin/env python3
# coding: utf-8
import caldav
import datetime
import icalendar
from . import calendar, getkey
from urllib import request


class Cdav:
    def __init__(self, caldav_dict, screen, conf):
        self._list = caldav_dict
        self._conf = conf
        self._icals = []        # list of Icalendar, for each caldav
                                # proper url if webcal; None if offline
        self._iscaldav = []     # list of True/False
        self._tags = []         # associated tags
        self._cal = []          # single list of all events in Agenda class
        self._cals = []         # list of event lists, by caldav
        self._screen = screen   # to give feedbacks in UI
        self._ymax, self._xmax = self._screen.getmaxyx()
        self.initialize()       # get cals and associated tags
        self.sync()             # update and redefine _events

    @property
    def icals(self):
        return self._icals

    @property
    def events(self):
        return self._cal.events

    def resize(self):
        self._ymax, self._xmax = self._screen.getmaxyx()

    def initialize(self):
        for cdav_idx, cdav in enumerate(self._list):
            self.initialize_single(cdav, cdav_idx)

    def initialize_single(self, cdav, cdav_idx):
        try:
            if "@" in cdav["url"]:  # there is a password in the url
                head_user_password, tail = cdav["url"].split("@")
                head, user_password = head_user_password.split("://")
                user = user_password.split(":")[0]
                url = head+"://"+user+":"+"*****"+"@"+tail
            else:
                url = cdav["url"]
            msg = _("Loading caldav ")+str(cdav_idx+1)+_(": ")+url
            self.inform(msg)

            if cdav["url"][:4] == "http":  # caldav
                flag = True  # it's a proper caldav
                client = self.get_client(cdav)
                principal = self.get_principal(client)
                self.add_calendar(principal)
            if cdav["url"][:6] == "webcal":  # webcal
                flag = False  # it's a webcal
                if cdav["url"][:7] == "webcals":  # webcal
                    urlh = "https" + cdav["url"][7:]
                else:
                    urlh = "http" + cdav["url"][6:]
                cal = read_webcal_url(urlh)
                self._icals.append(urlh)
            self._iscaldav.append(flag)
        except:
            self.resize()  # if screen sized changed during loading
            err = _("Error: can't contact caldav ")+cdav["url"]
            if calendar.load_local_caldav(cdav["url"], cdav["username"],
                                          self._conf._datapath, check=True):
                err += _(" -- using offline copy")
            err += _(" -- press any key to continue")
            self.inform(err)
            # append None to keep same size as _list
            self._icals.append(None)
            self._iscaldav.append(False)
            # pause; resize if necessary
            if getkey.getkey(self._screen) == "KEY_RESIZE":
                self.draw_conf_screen()
        self.add_tag(cdav)

    def get_client(self, cdav):
        if len(cdav["username"]) != 0:
            client = caldav.DAVClient(cdav["url"],
                                      username=cdav["username"],
                                      password=cdav["password"])
        else:
            client = caldav.DAVClient(cdav["url"])
        return client

    def get_principal(self, client):
        return client.principal()

    def add_calendar(self, principal):
        if len(principal.calendars()) == 0:
            cal = principal.make_calendar()
            # useful for my poor test of python-radicale
        else:
            # TODO: when several calendars, what to do?
            cal = principal.calendars()[0]
        self._icals.append(cal)

    def add_tag(self, cdav):
        if cdav:
            self._tags.append(cdav["tag"])
        else:
            self._tags.append(None)

    def sync(self):
        # extracts all the events of the the ICalendar
        # set all events in a single culendar/Agenda class
        self.inform(_("Synchronising caldavs…"))
        # resets the differents cals
        self._cals = []
        for ind, cal in enumerate(self._icals):
            tmpcal = calendar.Agenda()
            if self._iscaldav[ind]:
                ical = cal.events()
                for e in ical:  # a list of ical events, each one in a calendar
                    agenda = icalendar.Calendar()
                    agenda = agenda.from_ical(e.data)
                    agendacul = calendar.Agenda2Cul(agenda, [],
                                                    self._tags[ind],
                                                    cal)
                    ecul = agendacul.events[0]  # a single event
                    ecul.url = e.canonical_url
                    ecul.colour = self._list[ind]["colour"]
                    tmpcal.add_event(ecul)
            else:
                try:
                    # if unreachable, cal was None and thus, error
                    cal = read_webcal_url(cal)
                except:
                    pass
                # reached and a webcal
                if cal and cal[:15] == "BEGIN:VCALENDAR":
                    # in a webcal, _icals is url
                    # in a webcal, we get everything.
                    ical = icalendar.Calendar.from_ical(cal)
                    agendacul = calendar.Agenda2Cul(ical, [],
                                                    self._tags[ind],
                                                    "webcal")
                    for e in agendacul.events:
                        e.colour = self._list[ind]["colour"]
                        tmpcal.add_event(e)
                else:  # unreachable, use offline copy if any
                    url = self._list[ind]["url"]
                    username = self._list[ind]["username"]
                    tmpcal = calendar.load_local_caldav(url, username,
                                                        self._conf._datapath)
                    if cal:  # we got something, but not a wecal
                        warn = _("Warning: contacting ")
                        if len(url) > 20:
                            warn += url[:20] + "[…]"
                        else:
                            warn += url
                        warn += _(" gave a ")
                        warn += cal[:23]
                        warn += _(" -- using offline copy")
                        warn += _(" -- press any key to continue")
                        self.inform(warn)
                        if getkey.getkey(self._screen) == "KEY_RESIZE":
                            self.draw_conf_screen()
            if tmpcal:  # avoid None from loading offline apts
                tmpcal.sort()  # sort for better offline text copy
                self._cals.append(tmpcal)
        self._cal = calendar.Agenda()
        for cal in self._cals:
            for e in cal.events:
                self._cal.add_event(e)

        calendar.save_caldav(self._conf, self)  # local save

    def inform(self, msg):
        # clear footer
        self._screen.hline(self._ymax-1, 0, " ", self._xmax)
        self._screen.addstr(self._ymax-1, 0, msg[0:self._xmax-1])
        self._screen.refresh()

    def del_caldav(self, list_idx):
        # _list already removed in confscreen
        # clean what is required:
        # from highest position to lowest to pop easily
        list_idx.sort(reverse=True)
        for idx in list_idx:
            del_tag = self._tags.pop(idx)
            self._icals.pop(idx)
            self._iscaldav.pop(idx)

    def update(self, oldconf):
        list_idx = []
        for idx, conf in enumerate(oldconf):
            if conf not in self._list:
                list_idx.append(idx)
        if list_idx:
            self.del_caldav(list_idx)
        idx = 0
        for conf in self._list:
            if conf not in oldconf:
                self.initialize_single(conf, idx)
                idx += 1
        self.sync()


def del_event(event):
    event.caldav.event_by_url(event.url).delete()


def add_event(event, cdav):
    event.caldav = cdav
    ical = icalendar.Calendar()
    icalevent = calendar.event2icalevent(event)
    ical.add_component(icalevent)
    tmpevent = cdav.add_event(ical.to_ical().decode())
    event.url = tmpevent.canonical_url


def read_webcal_url(url):
    if "@" not in url:  # no password
        return request.urlopen(url).read().decode()
    else:
        # keep between http[s] *://* user:pwd *@* url
        user_and_password = url.split("@")[0].split("://")[1]
        user = user_and_password.split(":")[0]
        # if there are ":" in the password, join them again
        password = ":".join(user_and_password.split(":")[1:])
        # remove login and password from url
        url_clean = "".join(url.split(user+":"+password+"@"))

        # create a password manager and an opener
        pwd_mgr = request.HTTPPasswordMgrWithDefaultRealm()
        pwd_mgr.add_password(None, url_clean, user, password)
        handler = request.HTTPBasicAuthHandler(pwd_mgr)
        opener = request.build_opener(handler)

        # finally read and decode the data
        return opener.open(url_clean).read().decode()



#! /usr/bin/env python3
# coding: utf-8
import curses


def clicked(cul):
    try:
        _, x, y, _, click_type = curses.getmouse()
    except:
        x = -1
        y = -1
        click_type = -1
        click = "bug"
        cul.screen.addstr(1, 1, "bug detected")

    # the todo list has been clicked
    if cul.conf.todo and x >= cul._xmax:  # MOCHE²
        if not(cul._todohl):  # MOCHE
            cul.toggle_TODO()
        if (y-2) in range(len(cul._todo.events)):
            # select a todo, starting on line 2, todohl starts at 1
            cul._todohl = y-2 + 1
            cul.draw_todo()

    elif x <= cul._xoffset:
        # clicked on the hour? Why?
        pass
    else:  # clicked on the main culendar
        if cul._todohl:  # todo list is highlighted
            cul._todohl = 0
            cul.draw_todo()  # not anymore

        # select the current day
        day_lines = cul._lines[0]
        iday = 0
        for i in range(len(day_lines)):
            if x > day_lines[i]:
                iday += 1
        # Monday is 1
        cur_iday = cul.day.isoweekday()
        cul.addday(iday - cur_iday)

        # have we clicked on an event?
        calday = cul._calweek[iday-1]  # MOCHE
        # Monday is 1, first value is 0
        for i, e in enumerate(calday.events):
            if y in e.vlines and x in e.hlines:
                cul._ehl = i  # MOCHE
        cul.draw_day()



#! /usr/bin/env python3
# coding: utf-8
import curses
from . import getkey


class Help:
    def __init__(self, scr, conf):
        self._screen = scr
        self._conf = conf
        # help categories
        self._moving = ['nextday', 'prevday', 'nextweek', 'prevweek',
                        'nextmonth', 'prevmonth', 'nextyear', 'prevyear',
                        'today', 'startweek', 'endweek', 'setday']
        self._events = ['nextevent', 'prevevent', 'addevent',
                        'delevent',  'editevent', 'tagevent', 'copyevent',
                        'minusshifthour', 'shifthour', 'minusshiftday',
                        'shiftday']
        self._io = ['save', 'import', 'export']
        self._misc = ['toggleWE', 'redraw', 'sync',
                      'setconfig', 'help', 'quit']
        self._categories = [[self._moving, _("Choosing date:")],
                            [self._events, _("Dealing with events:")],
                            [self._io, _("Reading and writing files:")],
                            [self._misc, _("Miscellaneous:")]]

        # descriptions of all possibilities
        self._desc = {}
        self._desc['nextday'] = _('Select next day')
        self._desc['prevday'] = _('Select previous day')
        self._desc['nextweek'] = _('Go to next week')
        self._desc['prevweek'] = _('Go to previous week')
        self._desc['nextmonth'] = _('Go to next month')
        self._desc['prevmonth'] = _('Go to previous month')
        self._desc['nextyear'] = _('Go to next year')
        self._desc['prevyear'] = _('Go to previous year')
        self._desc['today'] = _('Go to today')
        self._desc['startweek'] = _('Go to first day of current week')
        self._desc['endweek'] = _('Go to last day of current week')
        self._desc['setday'] = _('Enter the day to go')
        desc = _('Select next event of current day or next todo')
        self._desc['nextevent'] = desc
        desc = _('Select previous event of current day or previous todo')
        self._desc['prevevent'] = desc
        self._desc['addevent'] = _('Add an event or todo')
        self._desc['delevent'] = _('Delete selected event or todo (if any)')
        self._desc['editevent'] = _('Edit selected event todo (if any)')
        desc = _('Set a numerical [0-7] tag to selected event')
        self._desc['tagevent'] = desc
        self._desc['tagevent2'] = _(' Alphabetic tag puts event on CalDAV')
        desc = _('Copy selected event to next day, week, month or year')
        self._desc['copyevent'] = desc
        desc = _("Shift selected event one hour earlier")
        self._desc['minusshifthour'] = desc
        self._desc['shifthour'] = _("Shift selected event one hour later")
        desc = _("Shift selected event one day earlier")
        self._desc['minusshiftday'] = desc
        self._desc['shiftday'] = _("Shift selected event one day later")
        self._desc['save'] = _('Save calendar and configuration')
        desc = _('Import a calendar (from iCalendar or calcurse)')
        self._desc['import'] = desc
        self._desc['export'] = _('Export calendar (to iCalendar or calcurse)')
        self._desc['toggleWE'] = _('Toggle presence of week-end')
        self._desc['toggletodo'] = _('Toggle presence of Todo list')
        self._desc['redraw'] = _('Redraw screen')
        self._desc['sync'] = _('Synchronize Caldavs and Webcals')
        self._desc['setconfig'] = _('Configure Culendar')
        self._desc['help'] = _('Get this help')
        self._desc['quit'] = _('Quit Culendar')

    def draw_help_screen(self):
        self._screen.clear()
        self._y, self._x = self._screen.getmaxyx()

        # if ridiculously small terminal
        if (self._x < 25) or (self._y < 5):
            self._screen.clear()
            text = _("\_x< A dead duck. That's all such a small terminal deserves.")
            self._screen.addstr(0, 0, text[0:(self._x-1)*(self._y-1)])

            key = ""
            while not((key in self._conf.keys['quit'])
                      or (key == 'KEY_RESIZE')):
                key = getkey.getkey(self._screen, self._conf.debug)
            if key in self._conf.keys['quit']:
                exit()
            else:
                self.draw_help_screen()

        # draw screen
        headleft = _("q: quit help")
        headcenter = _("Culendar help screen")
        headcenter += " 0.20230329"
        self._screen.border()
        self._screen.hline(2, 1, curses.ACS_HLINE, self._x-2)
        self._screen.addch(2, 0, curses.ACS_LTEE)
        self._screen.addch(2, self._x-1, curses.ACS_RTEE)
        self._screen.addstr(1, 1, headleft, curses.A_BOLD)
        self._screen.addstr(1, (self._x-len(headcenter))//2+1,
                            headcenter, curses.A_BOLD)

        # create full pad for self._screenolling help
        # counting lines for each category
        self._padlines = (3 + len(self._moving)+4 + len(self._events)+4
                          + len(self._io)+4 + len(self._misc)+4)
        colsize = round((self._x-2)/4)
        # pad columns: starting description + maximum description
        padcols = 2*colsize + max(len(v) for v in self._desc.values())
        self._pad = curses.newpad(self._padlines, padcols)
        self._pad.scrollok(True)
        self._screen.refresh()  # if not present, pad is ignored

        self._pad.addstr(0, 0, _("Key(s)"))
        self._pad.addstr(0, colsize, _("Function name"))
        self._pad.addstr(0, 2*colsize, _("Function description"))
        shift = 2
        for category in self._categories:
            # print category as subtitle
            self._pad.addstr(shift, 0, category[1], curses.A_BOLD)
            shift += 2

            if category[1] == _("Miscellaneous:"):
                # Repetition is not really a command
                # treat it separately
                self._pad.addstr(shift, 0, _("number n>0"))
                # second column: small name
                self._pad.addstr(shift, colsize, "repetition")
                # third column: description
                desc = _("Repeat following command n times (when reasonable)")
                self._pad.addstr(shift, 2*colsize, desc)
                shift +=1

            for i, k in enumerate(category[0]):
                # first column: keys
                listkeys = ""
                for ck in self._conf.keys[k]:
                    if ck == "\t":
                        listkeys += "TAB, "
                    elif ck == " ":
                        listkeys += "SPACE, "
                    else:
                        listkeys += ck+", "
                listkeys = listkeys[:-2]  # erase final comma
                self._pad.addstr(i+shift, 0, listkeys)
                # second column: small name
                self._pad.addstr(i+shift, colsize, k)
                # third column: description
                self._pad.addstr(i+shift, 2*colsize, self._desc[k])
                if k == "tagevent":
                    shift += 1
                    self._pad.addstr(i+shift, 2*colsize,
                                     self._desc["tagevent2"])
            shift += len(category[0]) + 2

        self._top = 0
        # the lines available for the pad
        self._padavail = self._y - 4
        # to have the bottom of the pad on the bottom of the screen
        self._maxtop = max(0, self._padlines - self._padavail)

    def help_screen(self):
        self.draw_help_screen()
        self._pad.refresh(self._top, 0, 3, 1, self._y-2, self._x-2)
        key = getkey.getkey(self._screen, self._conf.debug)
        while True:
            if key == 'KEY_RESIZE':
                self.draw_help_screen()
            if key == 'KEY_UP':
                self._top = max(0, self._top - 1)
            if key == 'KEY_DOWN':
                self._top = min(self._maxtop, self._top + 1)
            if key == 'KEY_PPAGE':
                self._top = max(0, self._top - self._padavail)
            if key == 'KEY_NPAGE':
                self._top = min(self._maxtop, self._top + self._padavail)
            if key in self._conf.keys['quit']:
                return
            self._pad.refresh(self._top, 0, 3, 1, self._y-2, self._x-2)
            key = getkey.getkey(self._screen, self._conf.debug)



#! /usr/bin/env python3
# coding: utf-8
import icalendar
import datetime
from os import path, rename


########################################################################
# Event class
########################################################################

class Event:
    def __init__(self, date, duration, summary, place="", tag=0,
                 url=None, caldav=None, colour=None):
        self._date = date          # datetime.datetime()
        self._duration = duration  # seconds
        self._summary = summary.strip()   # text
        self._place = place.strip()       # text
        self._tag = tag
        self._colour = colour  # default colour is tag; useful for caldav
        self._samehour = 0     # number of simultaneous events
        self._hlines = None    # horizontal lines of event
        self._vlines = None    # vertical lines of event
        # CalDAV
        self._url = url
        self._caldav = caldav

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, d):
        self._date = d

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, d):
        self._duration = d

    @property
    def summary(self):
        return self._summary

    @summary.setter
    def summary(self, s):
        self._summary = s

    @property
    def place(self):
        return self._place

    @place.setter
    def place(self, p):
        self._place = p

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, t):
        self._tag = t

    @property
    def samehour(self):
        return self._samehour

    @samehour.setter
    def samehour(self, samehour):
        self._samehour = samehour

    @property
    def hlines(self):
        return self._hlines

    @hlines.setter
    def hlines(self, hlines):
        self._hlines = hlines

    @property
    def vlines(self):
        return self._vlines

    @vlines.setter
    def vlines(self, vlines):
        self._vlines = vlines

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = url

    @property
    def caldav(self):
        return self._caldav

    @caldav.setter
    def caldav(self, caldav):
        self._caldav = caldav

    @property
    def colour(self):
        return self._colour

    @colour.setter
    def colour(self, colour):
        self._colour = colour

    def compute_vlines(self, cul):
        start_hour = int(self.date.strftime('%H'))
        start_min = int(self.date.strftime('%M'))
        end_time = start_hour*3600 + start_min*60 + self.duration
        # if starting hour doesn't fit, make it fit
        if start_hour < cul.hmin:
            start_hour = cul.hmin
            start_min = 0
        if start_hour >= cul.hmax:
            start_hour = cul.hmax
            start_min = 0
        if end_time < start_hour*3600:
            end_time = start_hour*3600

        # don't overflow
        end_time = min(cul.hmax*3600, end_time)
        end_hour = end_time//3600
        end_min = (end_time - end_hour*3600)//60

        # time to lines
        # select the lines
        startline = start_hour - cul.hmin
        endline = end_hour - cul.hmin
        shline = cul.lines[1][startline] + 1  # 1st line after hour line
        ehline = cul.lines[1][endline] + 1    # same as start hour

        # lines to range of line
        if cul.hour_size == 1:
            # small terminals: starts on line event
            shline = shline-1
            ehline = ehline-1
            pass
        else:  # tall terminal, work between lines, deal with minutes
            minute_size = round(60 / (cul.hour_size-1))
            if cul.hour_size > 2:
                # change start line only for wide hours
                shline += start_min // minute_size
            if end_min == 0:
                ehline -= 1  # stops before hour line
            else:
                ehline += end_min // minute_size

            # special case out of schedule
            if start_hour == cul.hmax:
                shline -= 1
        # failproof
        if ehline <= shline:
            ehline = shline+1

        # finally define the range
        self.vlines = range(shline, ehline)

    def compute_hlines(self, cul, col, width, maxevents, table):
        start_vline = table[self.vlines[0]]
        e_width = width//maxevents
        left = 0
        while sum(start_vline[left:(left+e_width)]) != 0:
            left += 1  # width//maxevents

        # have we gone too far?
            if left+e_width > len(start_vline):
                # reduce event width
                e_width = len(start_vline)-left

        # first column picked from lines
        for v in self.vlines:
            for i in range(e_width):
                table[v][left+i] = 1

        # first column: right of vertical day line + offset of multievent
        first_col = cul.lines[0][col] + 1 + left

        # finally define the range
        self.hlines = range(first_col, first_col + e_width)

        # too many events or too small terminal
        if len(self.hlines) < 1:
            self.hlines = (cul.lines[0][col]+1, width)


########################################################################
# Agenda class
########################################################################

class Agenda:
    def __init__(self):
        self._events = []

    def add_event(self, e):
        self._events.append(e)

    def del_event(self, e):
        self._events.remove(e)

    @property
    def events(self):
        return self._events

    # sort by date and time
    def sort(self):
        maxdate = datetime.date(9999, 12, 31)
        self._events = sorted(self._events,
                              key=lambda e: e.date if e.date else maxdate)


########################################################################
# other related functions
########################################################################

def extract_day(cal, caldav, day):
    # create a list of subcals for each day of the week
    calday = Agenda()

    for e in cal.events + caldav.calweek.events:
        # if the right day, add the event
        if e.date.toordinal() == day.toordinal():
            calday.add_event(e)

    calday.sort()
    return(calday)


def extract_week(cal, caldav, day, cul, autosync=1):
    # create a list of subcals for each day of the week
    calweek = []
    for i in range(7):
        calweek.append(Agenda())

    # selects the events in caldavs, if required
    if autosync:
        caldav.sync()

    # first day of the current week; +1 for starting on monday in ordinal
    firstday = day.toordinal() - day.isoweekday() + 1

    for e in cal.events + caldav.events:
        diffday = e.date.toordinal() - firstday
        # if in the right week, we have
        if 0 <= diffday <= 6:
            calweek[diffday].add_event(e)

    for iday in range(cul.daynb):  # for all printed days
        calweek[iday].sort()
        for e in calweek[iday].events:
            # compute each vertical position for events
            e.compute_vlines(cul)

        # initialize no event on all lines
        nbevent = [0 for h in range(cul.ymax)]
        for e in calweek[iday].events:
            for i in e.vlines:
                nbevent[i] += 1
        maxevent = max(nbevent)

        # if last idayumn, the day size is not cul.day_size
        if iday == cul.daynb-1:
            width = cul.lines[0][-1] - cul.lines[0][-2] - 1
        else:
            width = cul.day_size - 1
        # day_size - 1: doesn't include vertical separating lines

        # table of position of events on day
        table = [[0 for d in range(width)] for y in range(cul.ymax)]
        for e in calweek[iday].events:
            maxevent = max([nbevent[i] for i in e.vlines])
            e.compute_hlines(cul, iday, width, maxevent, table)

    return calweek


def import_ical(filename, categories, default_tag=0):
    with open(filename, "r") as f:
        ical = icalendar.Calendar.from_ical(f.read())
    return Agenda2Cul(ical, categories, default_tag)


def Agenda2Cul(agenda, categories, default_tag, caldav=None):
    cal = Agenda()
    for ie in agenda.walk('vevent'):
        sdate = ie['dtstart'].dt
        try:  # is there an end date?
            edate = ie['dtend'].dt
            # if so, compute duration
            dur = (edate - sdate).seconds
        except:
            try:  # there is thus a duration
                dur = ie['duration'].dt.seconds
            except:  # no end date, no duration… Are you kidding me?
                dur = 0  # it exists, that's all
        try:  # is there a summary?
            summ = ie['summary'].lstrip().replace('\n', "-")
        except:  # if none, NULL summary
            summ = " "
        try:  # is there a location?
            place = ie['location'].lstrip().replace('\n', "-")
        except:  # if none, NULL summary
            place = ""
        try:  # is there a category?
            cat = ie['category']
            tag = default_tag
            if type(cat) == list:  # several categories in a single event
                for c in cat:
                    for t in range(1, 8):
                        if c.lstrip() in categories[t]:
                            # if category exists, tag it
                            tag = t
            else:  # single category
                for t in range(1, 8):
                    if cat.lstrip() in categories[t]:
                        # if category exists, tag it
                        tag = t
        except:  # no category
            tag = default_tag
        # special case for events without hour:
        # set 1st hour to 0:00 and duration to 24h
        if type(sdate) == datetime.date:
            sdate = datetime.datetime.combine(sdate, datetime.time(0))
            dur = 24*60*60
        # check the potential timezone and put to local
        if sdate.tzinfo:
            sdate = sdate.astimezone()
            sdate = sdate.replace(tzinfo=None)
        if caldav:
            e = Event(sdate, dur, summ, place, tag, caldav=caldav)
        else:
            e = Event(sdate, dur, summ, place, tag)
        cal.add_event(e)
    return cal


def event2icalevent(e, categories=None):
    ie = icalendar.Event()
    ie.add('dtstart', e.date)
    ie.add('duration', datetime.timedelta(seconds=e.duration))
    ie.add('summary', e.summary)
    if e.place:
        ie.add('location', e.place)
    if e.tag and type(e.tag) == int:
        if len(categories[e.tag]) > 1:
            ie.add('category', categories[e.tag][1])
        else:
            ie.add('category', e.tag)
    if e.url:
        ie.add('url', e.url)
    return ie


def export_ical(cal, filename, categories):
    ical = icalendar.Calendar()
    ical.add('prodid', '-//From Culendar')
    ical.add('version', '2.0')

    for e in cal.events:
        ie = event2icalevent(e, categories)
        ical.add_component(ie)

    with open(filename, "wb") as f:
        f.write(ical.to_ical())


def import_calcurse(filename, t=0):
    cal = Agenda()
    with open(filename) as f:
        for line in f:
            smon = int(line[0:2])
            sday = int(line[3:5])
            syea = int(line[6:10])
            shou = int(line[13:15])
            smin = int(line[16:18])
            emon = int(line[22:24])
            eday = int(line[25:27])
            eyea = int(line[28:32])
            ehou = int(line[35:37])
            emin = int(line[38:40])
            summ = line[42:-1]  # strips final \n
            sdate = datetime.datetime(syea, smon, sday, shou, smin)
            edate = datetime.datetime(eyea, emon, eday, ehou, emin)
            e = Event(sdate, (edate-sdate).seconds, summ, tag=t)
            cal.add_event(e)
    return cal


def export_calcurse(cal, filename):
    lines = ""
    for e in cal.events:
        line = str(e.date.strftime('%m/%d/%Y @ %H:%M'))
        line += " -> "
        enddate = e.date + datetime.timedelta(seconds=e.duration)
        line += str(enddate.strftime('%m/%d/%Y @ %H:%M'))
        line += " |"
        line += e.summary
        if e.place:
            line += "@"
            line += e.place
        line += "\n"
        lines += line
    with open(filename, "w") as f:
        f.write(lines)


def load(filename, todofilename):
    return load_apts(filename), load_todo(todofilename)


def load_apts(filename, caldav=None):
    cal = Agenda()
    with open(filename) as f:
        for line in f:
            syea = int(line[0:4])
            smon = int(line[5:7])
            sday = int(line[8:10])
            shou = int(line[13:15])
            smin = int(line[16:18])
            eyea = int(line[22:26])
            emon = int(line[27:29])
            eday = int(line[30:32])
            ehou = int(line[35:37])
            emin = int(line[38:40])
            rema = line[42:-1]  # strips final \n
            # search for a tag
            try:
                if "|" in rema:
                    tag = int(rema[rema.find("|")+1:])
                    rema = rema[:rema.find("|")]
                else:
                    tag = 0
            except:
                tag = 0
            # search for a location delimiter
            if "@" in rema:
                summ = rema[:rema.find("@")]
                place = rema[rema.find("@")+1:]
            else:
                summ = rema
                place = ""
            sdate = datetime.datetime(syea, smon, sday, shou, smin)
            edate = datetime.datetime(eyea, emon, eday, ehou, emin)
            e = Event(sdate, (edate-sdate).seconds, summ, place, tag,
                      caldav=caldav)
            cal.add_event(e)
    cal.sort()
    return cal


def load_todo(todofilename):
    # Todo list: ugly hack, use Agenda
    # Todo item: Event with no duration and optional date
    todo = Agenda()
    with open(todofilename) as f:
        for line in f:
            try:  # if there is a date
                syea = int(line[0:4])
                smon = int(line[5:7])
                sday = int(line[8:10])
                sdate = datetime.date(syea, smon, sday)
                rema = line[14:-1]  # strips final \n
            except:
                rema = line[:-1]  # strips final \n
                sdate = None
            try:
                # if no |, empty tag and rema doesn't change
                tag = int(rema[rema.find("|")+1:])
                summ = rema[:rema.find("|")]
            except:
                tag = 0
                summ = rema
            e = Event(sdate, None, summary=summ, tag=tag)
            todo.add_event(e)
    todo.sort()
    return todo


def cal2txt(cal, colour=None):
    lines = ""
    for e in cal.events:
        line = str(e.date.strftime('%Y/%m/%d @ %H:%M'))
        line += " -> "
        enddate = e.date + datetime.timedelta(seconds=e.duration)
        line += str(enddate.strftime('%Y/%m/%d @ %H:%M'))
        line += " |"
        line += e.summary
        if e.place:
            line += "@"
            line += e.place
        if colour:
            line += "|"
            line += str(colour)
        elif e.tag:
            line += "|"
            line += str(e.tag)
        line += "\n"
        lines += line
    return lines


def todo2txt(todo):
    lines = ""
    for e in todo.events:
        if e.date:
            line = str(e.date.strftime('%Y/%m/%d'))
            line += " -> "
        else:
            line = ""
        line += e.summary
        if e.tag:
            line += "|"
            line += str(e.tag)
        line += "\n"
        lines += line
    return lines


def save(cal, todo, conf, caldav=None):
    # main Agenda
    # backup current one
    if path.exists(conf.datafile):
        rename(conf.datafile, conf.datafile+'.old')
    with open(conf.datafile, "w") as f:
        f.write(cal2txt(cal))

    # Todo file
    # backup current one
    if path.exists(conf.todofile):
        rename(conf.todofile, conf.todofile+'.old')

    with open(conf.todofile, "w") as f:
        f.write(todo2txt(todo))

    if caldav:
        save_caldav(conf, caldav)


def save_caldav(conf, caldav):
    # Caldav local saves - no backup, it's online
    for cdav_idx, cdav in enumerate(caldav._list):
        if caldav._icals[cdav_idx]:  # is there anything to save
            url = cdav["url"]
            suffix = url.replace("/", "-") + cdav["username"]
            fname = "apts." + suffix
            with open(conf._datapath+"/"+fname, "w") as f:
                f.write(cal2txt(caldav._cals[cdav_idx], cdav["colour"]))


def load_local_caldav(url, username, fpath, check=False):
    suffix = url.replace("/", "-") + username
    filename = fpath + "/" + "apts." + suffix
    if path.exists(filename):
        if check:  # just check existence
            return True
        else:
            return load_apts(filename, caldav="webcal")
    else:  # no local copy exists
        return None



#! /usr/bin/env python3
# coding: utf-8
import curses
import datetime
import icalendar
import sys
from .cul import config, culendar, calendar, helpscr, confscr, culdav, getkey
from .cul import mouse


def main(stdscr):
    # configialize ncurses and get conf
    conf = config.Config()

    # the current day
    day = datetime.date.today()
    if (day.isoweekday() in [6, 7]) & (conf.WE == 0):
        # if we are during an unshown day, go to the next Monday
        while day.isoweekday() in [6, 7]:
            day = datetime.date.fromordinal(day.toordinal() + 1)

    # get data and todo and caldavs urls
    cal, todo = calendar.load(conf.datafile, conf.todofile)
    cdav = culdav.Cdav(conf.caldav, stdscr, conf)

    # include first drawing
    cul = culendar.Culendar(stdscr, conf, day, cal, cdav, todo)

    # main loop
    key = ""
    while not(key in conf.keys['quit']):
        key = getkey.getkey(cul.screen, conf.debug)

        if key == 'KEY_RESIZE':
            cul.update()

        # repeat action at least 1, or the entered count value
        repetitions = max(1, cul.count)
        # changing day or week or year
        if key in conf.keys['prevday']:
            cul.addday(-repetitions)
        if key in conf.keys['nextday']:
            cul.addday(+repetitions)
        if key in conf.keys['nextweek']:
            cul.addday(+7*repetitions)
        if key in conf.keys['prevweek']:
            cul.addday(-7*repetitions)
        if key in conf.keys['nextmonth']:
            cul.addmonth(+repetitions)
        if key in conf.keys['prevmonth']:
            cul.addmonth(-repetitions)
        if key in conf.keys['nextyear']:
            cul.addyear(+repetitions)
        if key in conf.keys['prevyear']:
            cul.addyear(-repetitions)
        if key in conf.keys['today']:
            cul.changeday(day=datetime.date.today())
        if key in conf.keys['setday']:
            cul.changeday()
        if key in conf.keys['startweek']:
            cul.addday(-cul.day.isoweekday()+1)
        if key in conf.keys['endweek']:
            cul.addday(-cul.day.isoweekday()+cul.daynb)

        # change highlighted event of current day
        if key in conf.keys['nextevent']:
            cul.sameday()
            cul.changehl(+repetitions)
            cul.draw_hl()
        if key in conf.keys['prevevent']:
            cul.sameday()
            cul.changehl(-repetitions)
            cul.draw_hl()

        # delete, edit, add (highlighted) event (if existing)
        if key in conf.keys['delevent']:
            cul.del_event()
        if key in conf.keys['editevent']:
            cul.edit_event()
        if key in conf.keys['addevent']:
            cul.add_event()
        if key in conf.keys['tagevent']:
            cul.tag_event()
        if key in conf.keys['copyevent']:
            cul.copy_event(repetitions)
        if key in conf.keys['minusshifthour']:
            cul.shift_event(-repetitions)
        if key in conf.keys['minusshiftday']:
            cul.shift_event(-repetitions*24)
        if key in conf.keys['shifthour']:
            cul.shift_event(+repetitions)
        if key in conf.keys['shiftday']:
            cul.shift_event(+repetitions*24)

        # toggle week-end display
        if key in conf.keys['toggleWE']:
            cul.toggle_WE()

        # toggle Todo bar focus (appear/disappear)
        if key in conf.keys['toggletodo']:
            cul.toggle_TODO()

        if key in conf.keys['redraw']:
            cul.clear_cal()
            cul.draw_cal()
        if key in conf.keys['sync']:
            cul.sync_caldav()

        if key in conf.keys['help']:
            H = helpscr.Help(cul.screen, cul.conf)
            H.help_screen()
            # redraw, works even in case of KEY_RESIZE inside help
            cul.update()
        if key in conf.keys['save']:
            cul.inform(_("Saving configuration and calendar…"))
            conf.save()
            cul.save()
            cul.inform(_("Configuration and calendar saved!"))
        if key in conf.keys['setconfig']:
            C = confscr.Conf(cul.screen, cul.conf, cul.caldav)
            # real copy to keep a backup
            old_caldav_conf = [conf for conf in cul.conf._caldav]
            # conf_screen returns a possibly modified conf
            cul.conf = C.conf_screen()
            if old_caldav_conf != cul.conf._caldav:
                cul.caldav.update(old_caldav_conf)
            # redraw, works even in case of KEY_RESIZE inside help
            cul.update()
        if key in conf.keys['import']:
            cul.importcal()
        if key in conf.keys['export']:
            cul.exportcal()
        if key == "KEY_MOUSE":
            mouse.clicked(cul)

        # define the buffer and the repetition of actions
        try:  # is the key an integer?
            # type 1 then 2, get 12
            cul.count = cul.count*10 + int(key)
        except:  # not a number, back to zero
            cul.count = 0

    # autosave on quit
    if conf.autosave:
        cul.save()
        conf.save()


# here starts culendar!
def run():
    # TODO: parse arguments
    curses.wrapper(main)


if __name__ == "__main__":
    run()



#! /usr/bin/env python3
# coding: utf-8
import curses
import datetime
from . import getkey


# TODO: deal with too long lines
class Input:
    def __init__(self, screen, text="", pwd=False):
        self._screen = screen
        self.screen_update()
        self.text = text
        # self._maxpos is cursor max position in x, defined by text setter
        self._pos = self._maxpos   # cursor position in text
        self._pwd = pwd

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, newtext):
        self._text = newtext
        self._maxpos = len(self._text)

    def screen_update(self):
        self._ymax, self._xmax = self._screen.getmaxyx()
        self._y = self._ymax - 1

    def reset_text(self):
        self.text = ""
        self._pos = 0

    def delete_char(self, offset):
        if self._pos == 0:
            offset = 0  # avoid backspacing -1 character
        deleted = self._pos + offset
        self.text = self.text[:deleted] + self.text[deleted+1:]
        self._pos += offset
        # erase and redraw the new text
        self.delete_text()
        self.draw_text()

    def delete_part(self, begin, end):
        self.text = self.text[:begin] + self.text[end:]
        self._pos = max(0, self._pos - (end-begin))
        # erase and redraw the new text
        self.delete_text()
        self.draw_text()

    def add_char(self, key):
        self.text = self.text[:self._pos] + key + self.text[self._pos:]
        self._pos += 1
        self.draw_text()

    def draw_text(self):
        lentext = len(self._text)
        old_pos = None
        if lentext > self._xmax - 1:
            text = self._text[lentext - self._xmax + 1:]
            old_pos = self._pos
            self._pos = self._xmax - 1
        else:
            text = self._text
        if not self._pwd:
            self._screen.addstr(self._y, 0, text)
        else:  # obfuscate password
            self._screen.addstr(self._y, 0, "*"*len(text))
        self.draw_cursor()
        if old_pos:
            self._pos = old_pos

    def delete_text(self):
        self._screen.hline(self._y, 0, " ", self._xmax-1)

    def draw_cursor(self):
        # put the cursos at the good position
        if self._pos > self._xmax-1:
            self._screen.move(self._y, self._xmax-1)
        else:
            self._screen.move(self._y, self._pos)

    def move_cursor(self, newpos):
        self._pos = newpos
        self.draw_cursor()

    def delete_from_cursor(self, direction):
        old_maxpos = self._maxpos
        if direction == "left":
            self.text = self.text[self._pos:]
            self._pos = 0
        elif direction == "right":
            self.text = self.text[:self._pos]
            self._pos = self._maxpos
        # erase and redraw the new text
        self.delete_text()
        self.draw_text()

    def find_left_word(self):
        idx = -1
        for i, s in enumerate(self.text[:self._pos-1]):
            if s == " ":
                idx = i + 1
        if (idx > 0) & (idx < self._pos):  # something has been found
            # second condition to prevent the search from slice [:-1]
            return idx
        else:
            return 0

    def move_left_word(self):
        self._pos = self.find_left_word()
        self.draw_cursor()

    def delete_left_word(self):
        goal = self.find_left_word()
        self.delete_part(goal, self._pos)

    def find_right_word(self):
        idx = self.text[self._pos+1:].find(" ")
        if idx > 0:  # something has been found
            return self._pos + self.text[self._pos+1:].find(" ") + 1
        else:
            return self._maxpos

    def move_right_word(self):
        self._pos = self.find_right_word()
        self.draw_cursor()

########################################################################
# get input and definition of keybindings
########################################################################

    def get_input(self):
        curses.curs_set(1)  # start of input: redraw cursor
        key = 0  # integer that does nothing in the loop
        self.draw_text()
        while not(key in ["\n", ""]):
            if key in ["KEY_RESIZE"]:
                return -1

            elif key in ["KEY_DC", ""]:
                self.delete_char(0)

            elif key in ["KEY_IC"]:
                pass  # TODO

            elif key in ["KEY_BACKSPACE", ""]:
                self.delete_char(-1)

            elif key in ["KEY_HOME", "KEY_DOWN", ""]:
                self.move_cursor(0)

            elif key in ["KEY_END", "KEY_UP", ""]:
                self.move_cursor(self._maxpos)

            elif key in ["KEY_LEFT",  ""]:
                self.move_cursor(max(0, self._pos - 1))

            elif key in ["KEY_RIGHT", ""]:
                self.move_cursor(min(self._maxpos, self._pos + 1))

            elif key in [""]:
                self.delete_from_cursor("left")

            elif key in [""]:
                self.delete_from_cursor("right")

            elif key in ["KEY_SLEFT", "KEY_CLEFT"]:
                self.move_left_word()

            elif key in ["KEY_SRIGHT", "KEY_CRIGHT"]:
                self.move_right_word()

            elif key in [""]:
                self.delete_left_word()

            else:
                if type(key) == str:
                    self.add_char(key)

            key = getkey.getkey(self._screen)

        curses.curs_set(0)  # end of input: remove cursor

        if key == "":  # escape key: cancel text
            return 1
        else:
            return 0

########################################################################
# Check input validity
########################################################################
    def error(self, errmsg):
        self._screen.hline(self._ymax-2, 0, " ", self._xmax)
        self._screen.hline(self._ymax-1, 0, " ", self._xmax)
        self._screen.addstr(self._ymax-2, 0, errmsg)
        self._screen.addstr(self._ymax-1, 0,
                            _("Press [Enter] to continue"))
        k = getkey.getkey(self._screen)
        if k in ['q', 'Q', '']:
            return "CUL_CANCEL"
        elif k == 'KEY_RESIZE':
            return k
        else:
            return None

    def check_hour(self, day, start=False):
        try:  # to create a time
            if ":" in self.text:
                h = int(self.text[:self.text.find(":")])
                m = int(self.text[self.text.find(":")+1:])
            elif len(self.text) < 3:
                h = int(self.text)
                m = 0
            else:
                m = int(self.text[-2:])
                h = int(self.text[:-2])
            if h == 24:
                if start:
                    h = 23
                    m = 59
                    tmpday = day
                else:
                    h = 0
                    # add hour and minute to day by using datetime.datetime
                    tmpday = datetime.datetime.fromordinal(day.toordinal()+1)
            else:  # idem
                tmpday = datetime.datetime.fromordinal(day.toordinal())
            d = tmpday.replace(hour=h, minute=m)
            return d
        except:
            errmsg = _("You entered an invalid time")
            return self.error(errmsg)

    def check_duration(self, day, start_date):
        if "+" in self.text:
            end_date = self.check_add_time(start_date)
        else:
            end_date = self.check_hour(day)
        if not(end_date in [None, "CUL_CANCEL", "KEY_RESIZE"]):
            if end_date < start_date:
                st = (str(start_date.hour) + ":"
                      + ("0"+str(start_date.minute))[-2:])
                errmsg = _("Ending time should be greater than ")+st
                return self.error(errmsg)
        return end_date

    def check_add_time(self, start_date):
        try:
            duration = int(self.text[1:])*60  # in seconds
            end_date = start_date + datetime.timedelta(seconds=duration)
            return end_date
        except:
            errmsg = _("You entered an invalid duration")
            return self.error(errmsg)

    def check_day(self, day):
        if not(self.text):  # empty day: set to today
            return datetime.date.today()
        else:
            try:
                if '/' in self.text:
                    if len(self.text) > 5:
                        y = int(self.text[0:4])
                        dec = 5
                    else:
                        y = day.year
                        dec = 0
                    m = int(self.text[dec:dec+2])
                    d = int(self.text[dec+3:])
                else:
                    if len(self.text) > 4:
                        y = int(self.text[0:4])
                        dec = 4
                    else:
                        y = day.year
                        dec = 0
                    m = int(self.text[dec:dec+2])
                    d = int(self.text[dec+2:])
                return datetime.date(y, m, d)
            except:
                errmsg = _("Error: could not understand date")
                return self.error(errmsg)

    def check_filename(self, mode):
        try:
            if ((self.text[-2] == " ")
               and (self.text[-1] in [str(i) for i in range(8)])):
                # special case: hidden final tag
                f = open(self.text[:-2], mode)
                f.close()
            else:
                f = open(self.text, mode)
                f.close()
            return self.text
        except:
            if mode == "rb":
                errmsg = _("Error: unable to read the file ")
            elif mode == "wb":
                errmsg = _("Error: unable to write the file ")
            return self.error(errmsg + self.text)

    def check_inthour(self):
        try:
            if int(self.text) < 0:
                errmsg = _("Error: should be higher than 0")
                return self.error(errmsg)
            if int(self.text) > 24:
                errmsg = _("Error: should be lower than 24")
                return self.error(errmsg)
            return int(self.text)
        except:
            errmsg = _("Error: must be an integer")
            return self.error(errmsg)



#! /usr/bin/env python3
# coding: utf-8

import curses
import locale
import gettext
from os import environ, getenv, path, mkdir, rename


class Config:
    def __init__(self):
        # Mouse activation
        self._mmask = (curses.BUTTON1_CLICKED + curses.BUTTON1_DOUBLE_CLICKED
                       + curses.BUTTON1_TRIPLE_CLICKED)
        # invisible cursor
        curses.curs_set(0)
        # allow transparency
        curses.use_default_colors()
        # use the locale!
        locale.setlocale(locale.LC_ALL, '')
        gettext.install("culendar",
                        path.dirname(path.realpath(__file__))+"/../locale")

        # create default values
        self._hmin = 7   # 7 am
        self._hmax = 20  # 8 pm
        self._WE = bool(1)  # show week-ends
        self._todo = bool(0)      # doesn't show todo list
        self._todowidth = 0.2     # defaults 20% width
        self._autosave = bool(1)  # defaults to be less unusual
        self._colourset = 0       # defaults to flashy
        self._caldav = []         # empty list of caldavs
        self._autosync = False    # autoupdate all caldavs when changing day
        self._mouse = False       # disable mouse
        curses.mousemask(0)
        self._debug = bool(0)     # you want to see dead ducks

        self.checkdir()  # define paths
        self._configfile = self._configpath + '/culendar.conf'
        self._keysfile = self._configpath + '/keys.conf'
        self._coloursfile = self._configpath + '/colours.conf'
        self._categoriesfile = self._configpath + '/categories.conf'
        self._caldavfile = self._configpath + '/caldav.conf'
        self._datafile = self._datapath + '/apts'
        self._todofile = self._datapath + '/todo'
        # create blanks one if not existing
        if not(path.exists(self._datafile)):
            with open(self._datafile, "w") as f:
                f.write("")
        if not(path.exists(self._todofile)):
            with open(self._todofile, "w") as f:
                f.write("")

        self.readconf()         # overwrites default values
        self.readkeys()         # defines self._keys
        self.readcolours()      # defines self._colours_idx
        self.setcolours()       # defines self._colours
        self.readcategories()   # defines self._categories
        self.readcaldav()       # get the caldavs configurations

    @property
    def hmin(self):
        return self._hmin

    @hmin.setter
    def hmin(self, hmin):
        if hmin >= self._hmax:
            self._hmin = self._hmax-1
        else:
            self._hmin = max(0, hmin)

    @property
    def hmax(self):
        return self._hmax

    @hmax.setter
    def hmax(self, hmax):
        if hmax <= self._hmin:
            self._hmax = self._hmin+1
        else:
            self._hmax = min(24, hmax)

    @property
    def WE(self):
        return self._WE

    @WE.setter
    def WE(self, WE):
        self._WE = WE

    @property
    def todo(self):
        return self._todo

    @todo.setter
    def todo(self, T):
        self._todo = T

    @property
    def todowidth(self):
        return self._todowidth

    @property
    def autosave(self):
        return self._autosave

    @autosave.setter
    def autosave(self, autosave):
        self._autosave = autosave

    @property
    def keys(self):
        return self._keys

    @keys.setter
    def keys(self, keys):
        self._keys = keys

    @property
    def datafile(self):
        return self._datafile

    @property
    def todofile(self):
        return self._todofile

    @property
    def colourset(self):
        return self._colourset

    @colourset.setter
    def colourset(self, cs):
        self._colourset = cs

    @property
    def colours_idx(self):
        return self._colours_idx

    @colours_idx.setter
    def colours_idx(self, idx):
        self._colours_idx = idx

    @property
    def colours(self):
        return self._colours

    @colours.setter
    def colours(self, colours):
        self._colours = colours

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, categories):
        self._categories = categories

    @property
    def caldav(self):
        return self._caldav

    @caldav.setter
    def caldav(self, caldav):
        self._caldav = caldav

    @property
    def autosync(self):
        return self._autosync

    @autosync.setter
    def autosync(self, autosync):
        self._autosync = autosync

    @property
    def mouse(self):
        return self._mouse

    @mouse.setter
    def mouse(self, mouse):
        self._mouse = mouse
        if mouse:
            curses.mousemask(self._mmask)
        else:
            curses.mousemask(0)

    @property
    def debug(self):
        return self._debug

    def checkdir(self):
        # configuration stuff
        if 'XDG_CONFIG_HOME' in environ:
            self._configpath = getenv('XDG_DATA_HOME')
        else:
            self._configpath = getenv('HOME')+'/.config/culendar'

        if not(path.exists(self._configpath)):
            # implicitly suppose that $HOME/.config already exists
            mkdir(self._configpath)

        # data stuff
        if 'XDG_DATA_HOME' in environ:
            self._datapath = getenv('XDG_DATA_HOME')
        else:
            self._datapath = getenv('HOME')+'/.local/share/culendar'

        if not(path.exists(self._datapath)):
            # implicitly suppose that $HOME/.local/share already exists
            mkdir(self._datapath)

    def readconf(self):
        if not(path.exists(self._configfile)):
            # write the default conf
            self.writeconf()
        else:
            with open(self._configfile) as f:
                for line in f:
                    if line[0:7] == 'hmin = ':
                        # don't use the setter to have no limit on hmin
                        self._hmin = int(line[7:9])
                    if line[0:7] == 'hmax = ':
                        # use the setter to be sure to be higher than hmin
                        self.hmax = int(line[7:9])
                    if line[0:5] == 'WE = ':
                        self._WE = bool(int(line[5]))  # "0" to 0 to False
                    if line[0:11] == 'autosave = ':
                        self._autosave = bool(int(line[11]))
                    if line[0:12] == 'colourset = ':
                        self._colourset = int(line[12])
                    if line[0:11] == 'autosync = ':
                        self._autosync = bool(int(line[11]))
                    if line[0:8] == 'debug = ':
                        self._debug = bool(int(line[8]))
                    if line[0:7] == 'todo = ':
                        self._todo = bool(int(line[7]))
                    if line[0:12] == 'todowidth = ':
                        self._todowidth = float(line[12:])
                    if line[0:8] == 'mouse = ':
                        self.mouse = bool(int(line[8:]))
                # TODO: check if all is fine

    def writeconf(self):
        with open(self._configfile, "w") as f:
            f.write('hmin = ' + str(self._hmin) + '\n')
            f.write('hmax = ' + str(self._hmax) + '\n')
            f.write('WE = ' + str(int(self._WE)) + '\n')
            f.write('autosave = ' + str(int(self._autosave)) + '\n')
            f.write('colourset = ' + str(self._colourset) + '\n')
            f.write('autosync = ' + str(int(self._autosync))+'\n')
            f.write('todo = ' + str(int(self._todo))+'\n')
            f.write('todowidth = ' + str(self._todowidth)+'\n')
            f.write('mouse = ' + str(int(self._mouse)))
            if self._debug:  # if the debug is forced, write it
                f.write('\ndebug = ' + str(int(self._debug)))

    def readkeys(self):
        self._keys = {}
        if path.exists(self._keysfile):
            with open(self._keysfile) as f:
                for line in f:
                    # find separator between keyword and keys
                    ind = line.find(":")
                    keyword = line[:ind]
                    # +1 ignore :, -1 ignore final \n
                    keys = line[ind+1:-1].replace(" ", "").split(",")
                    if "SPACE" in keys:
                        keys.append(" ")
                        keys.remove("SPACE")
                    self._keys[keyword] = keys

        # check if all operations are defined, defaults if needed
        oldkeys = self._keys.copy()
        try:
            self._keys['quit']
        except:
            self._keys['quit'] = ['q', 'Q']
        try:
            self._keys['nextday']
        except:
            self._keys['nextday'] = ['KEY_RIGHT']
        try:
            self._keys['prevday']
        except:
            self._keys['prevday'] = ['KEY_LEFT']
        try:
            self._keys['nextweek']
        except:
            self._keys['nextweek'] = ['KEY_DOWN']
        try:
            self._keys['prevweek']
        except:
            self._keys['prevweek'] = ['KEY_UP']
        try:
            self._keys['nextmonth']
        except:
            self._keys['nextmonth'] = ['m']
        try:
            self._keys['prevmonth']
        except:
            self._keys['prevmonth'] = ['M']
        try:
            self._keys['nextyear']
        except:
            self._keys['nextyear'] = ['KEY_PPAGE', 'y']
        try:
            self._keys['prevyear']
        except:
            self._keys['prevyear'] = ['KEY_NPAGE', 'Y']
        try:
            self._keys['nextevent']
        except:
            self._keys['nextevent'] = ['KEY_END', '	']  # tab
        try:
            self._keys['prevevent']
        except:
            self._keys['prevevent'] = ['KEY_HOME', 'KEY_BTAB']
        try:
            self._keys['delevent']
        except:
            self._keys['delevent'] = ['d', 'D']
        try:
            self._keys['addevent']
        except:
            self._keys['addevent'] = ['a', 'A']
        try:
            self._keys['editevent']
        except:
            self._keys['editevent'] = ['e', 'E']
        try:
            self._keys['toggleWE']
        except:
            self._keys['toggleWE'] = ['w', 'W']
        try:
            self._keys['redraw']
        except:
            self._keys['redraw'] = ['', '']
        try:
            self._keys['sync']
        except:
            self._keys['sync'] = ['']
        try:
            self._keys['save']
        except:
            self._keys['save'] = ['s', 'S']
        try:
            self._keys['import']
        except:
            self._keys['import'] = ['i', 'I']
        try:
            self._keys['export']
        except:
            self._keys['export'] = ['x', 'X']
        try:
            self._keys['today']
        except:
            self._keys['today'] = ['']
        try:
            self._keys['setday']
        except:
            self._keys['setday'] = ['g', 'G']
        try:
            self._keys['startweek']
        except:
            self._keys['startweek'] = ['0']
        try:
            self._keys['endweek']
        except:
            self._keys['endweek'] = ['$']
        try:
            self._keys['help']
        except:
            self._keys['help'] = ['h', 'H']
        try:
            self._keys['tagevent']
        except:
            self._keys['tagevent'] = ['t', 'T', '|']
        try:
            self._keys['copyevent']
        except:
            self._keys['copyevent'] = ['p', 'P']
        try:
            self._keys['minusshifthour']
        except:
            self._keys['minusshifthour'] = ['-']
        try:
            self._keys['shifthour']
        except:
            self._keys['shifthour'] = ['+']
        try:
            self._keys['minusshiftday']
        except:
            self._keys['minusshiftday'] = ['/']
        try:
            self._keys['shiftday']
        except:
            self._keys['shiftday'] = ['*']
        try:
            self._keys['setconfig']
        except:
            self._keys['setconfig'] = ['c', 'C']
        try:
            self._keys['toggletodo']
        except:
            self._keys['toggletodo'] = ['']

        if len(self._keys) > len(oldkeys):
            # if a default configuration has been used
            self.writekeys()

    def writekeys(self):
        lines = ""
        # ordered list of items
        keywords = ['nextday', 'prevday', 'nextweek', 'prevweek',
                    'nextmonth', 'prevmonth', 'nextyear', 'prevyear',
                    'nextevent', 'prevevent', 'delevent', 'addevent',
                    'editevent', 'toggleWE', 'quit', 'redraw', 'sync', 'save',
                    'import', 'export', 'today', 'setday', 'startweek',
                    'endweek', 'help', 'tagevent', 'copyevent',
                    'minusshifthour', 'shifthour', 'minusshiftday', 'shiftday',
                    'setconfig', 'toggletodo']
        for op in keywords:
            line = op + ": "
            for k in self._keys[op]:
                if k == " ":
                    k = "SPACE"
                line = line + k + ", "
            # replace last comma and space by a newline
            line = line[:-2] + "\n"
            lines += line
        with open(self._keysfile, "w") as f:
            f.write(lines)

    def readcolours(self):
        if not(path.exists(self._coloursfile)):
            # defaults and write file
            self._colours_idx = range(8)
            self.writecolours()
        else:
            self._colours_idx = []
            with open(self._coloursfile) as f:
                for line in f:
                    self._colours_idx.append(int(line))

    def writecolours(self):
        lines = ""
        for colour_idx in self._colours_idx:
            lines += str(colour_idx) + "\n"
        with open(self._coloursfile, "w") as f:
            f.write(lines)

    def setcolours(self):
        self._colours = []
        # define the eight colours
        for i in range(8):
            try:  # the terminal has a default background
                curses.init_pair(i, i, -1)
            except:  # defaults to black
                curses.init_pair(i, i, 0)
        for idx in self._colours_idx:
            self._colours.append(curses.color_pair(idx))

    def readcategories(self):
        self._categories = {}
        if not(path.exists(self._categoriesfile)):
            # defaults and write file
            for i in range(1, 8):
                self._categories[i] = [str(i)]
            self.writecategories()
        else:
            with open(self._categoriesfile) as f:
                for idx, line in enumerate(f):
                    cat = line[:-1].replace(", ", ",").split(",")
                    # -1 to ignore final \n
                    self._categories[idx+1] = cat
                    # +1: category 0 is defaults

    def writecategories(self):
        lines = ""
        for tagnb in range(1, 8):
            line = ""
            for cat in self._categories[tagnb]:
                line = line + cat + ", "
            # replace last comma and space by a newline
            line = line[:-2] + "\n"
            lines += line
        with open(self._categoriesfile, "w") as f:
            f.write(lines)

    def readcaldav(self):
        if path.exists(self._caldavfile):
            with open(self._caldavfile) as f:
                for line in f:
                    cdav = {}  # initiate the dict
                    endurl = line.find(", username=")
                    endusername = line.find(", password=")
                    endpassword = line.find(", tag=")
                    endtag = line.find(", colour=")
                    # 5 to skip the starting  'url="'
                    url = line[5:endurl-1]
                    # is there a password?
                    if endusername - endurl > 13:
                        username = line[endurl+12:endusername-1]
                        password = line[endusername+12:endpassword-1]
                    else:
                        username = ""
                        password = ""
                    tag = line[endpassword+7:endtag-1]
                    # line finish by 'colour="X"\n', get the X
                    colour = int(line[-3])
                    # fill up the caldav's dict
                    cdav["url"] = url
                    cdav["username"] = username
                    cdav["password"] = password
                    cdav["tag"] = tag
                    cdav["colour"] = colour
                    # stack it in the list
                    self._caldav.append(cdav)

    def writecaldav(self):
        lines = ""
        for cdav in self._caldav:
            line = ""
            line += 'url="' + cdav["url"]
            line += '", username="' + cdav["username"]
            line += '", password="' + cdav["password"]
            line += '", tag="' + cdav["tag"]
            line += '", colour="' + str(cdav["colour"]) + '"\n'
            lines += line
        with open(self._caldavfile, "w") as f:
            f.write(lines)

    def save(self):
        self.writeconf()
        self.writekeys()
        self.writecolours()
        self.writecategories()
        self.writecaldav()
